@@ -28,11 +28,12 @@
 int always_readline = FALSE;	             /* -a option: always be in readline mode             */
 char *password_prompt_search_string = NULL;  /* (part of) password prompt (argument of -a option) */
 int ansi_colour_aware = FALSE;               /* -A option: make readline aware of ANSI colour codes in prompt */
+int bleach_the_prompt = FALSE;               /* -A!: remove all ANSI colour codes in prompt      */
 int complete_filenames = FALSE;	             /* -c option: whether to complete file names        */
 int debug = 0;			             /* -d option: debugging mask                        */
 char *extra_char_after_completion = " ";     /* -e option: override readlines's default completion_append_char (space) */
-int history_duplicate_avoidance_policy =
-  ELIMINATE_SUCCESIVE_DOUBLES;               /* -D option: whether and how to avoid duplicate history entries */
+int always_echo = FALSE;                     /* -E option: always echo, even if client has ECHO off */
+int history_duplicate_avoidance_policy =  ELIMINATE_SUCCESIVE_DOUBLES;               /* -D option: whether and how to avoid duplicate history entries */
 char *history_format = NULL;                 /* -F option: format to append to history entries            */
 char *forget_regexp = NULL;                  /* -g option: keep matching input out of history           */
 int pass_on_sigINT_as_sigTERM =  FALSE;      /* -I option: send a SIGTERM to client when a SIGINT is received */
@@ -41,6 +42,7 @@ int nowarn = FALSE;		             /* -n option: suppress warnings */
 int commands_children_not_wrapped =  FALSE;  /* -N option: always use direct mode when <command> is waiting */
 int one_shot_rlwrap = FALSE;                 /* -o option: whether to close the pty after writing the first line to <command> */
 char *prompt_regexp = NULL;		     /* -O option: only ever "cook" prompts matching this regexp */
+bool regexp_means_prompt = FALSE;            /* -O! all candidate prompts that match rhe regexp are prompts */
 int colour_the_prompt = FALSE;	             /* -p option: whether we should paint the prompt */
 int renice = FALSE;                          /* -R option: whether to be nicer than command */
 int mirror_arguments = FALSE;                /* -U option: whether to mirror command's arguments */
@@ -52,27 +54,30 @@ char *filter_command = NULL;                 /* -z option: pipe prompts, input,
 
 
 /* variables for global bookkeeping */
-int master_pty_fd;		     /* master pty (rlwrap uses this to communicate with client) */
-FILE *debug_fp = NULL;  	     /* filehandle of debugging log */
-char *program_name, *command_name;   /* "rlwrap" (or whatever has been symlinked to rlwrap)  and (base-)name of command */
-char *rlwrap_command_line = "";      /* rlwrap command line (rlwrap -options command <command_args> */
-char *command_line = "";             /* command <command_args> */
-int within_line_edit = FALSE;	     /* TRUE while user is editing input */
-pid_t command_pid = 0;		     /* pid of child (client), or 0 before child is born */
-int i_am_child = FALSE;		     /* Am I child or parent? after forking, child will set this to TRUE */
-int ignore_queued_input = FALSE;     /* read and then ignore all characters in input queue until it is empty (i.e. read would block) */
-int received_WINCH = FALSE;          /* flag set in SIGWINCH signal handler: start line edit as soon as possible */
-int prompt_is_still_uncooked = TRUE; /* The main loop consults this variable to determine the select() timeout: when TRUE, it is
-                                        a few millisecs, if FALSE, it is infinite.						
-                                        TRUE just after receiving command output (when we still don't know whether we have a
-                                        prompt), and, importantly, at startup (so that substitute prompts get displayed even with
-                                        programs that don't have a startup message, such as cat)  */
+int master_pty_fd;		             /* master pty (rlwrap uses this to communicate with client) */
+FILE *debug_fp = NULL;  	             /* filehandle of debugging log */
+char *program_name = "rlwrap", *command_name;/* "rlwrap" (or whatever has been symlinked to rlwrap) and (base-)name of command */
+char *rlwrap_command_line = "";              /* rlwrap command line (rlwrap -options command <command_args> */
+char *command_line = "";                     /* command <command_args> */
+int within_line_edit = FALSE;	             /* TRUE while user is editing input */
+int screen_is_alternate = FALSE;             /* TRUE after client has sent smcup, FALSE after rmcup */
+pid_t command_pid = 0;		             /* pid of child (client), or 0 before child is born */
+int i_am_child = FALSE;		             /* Am I child or parent? after forking, child will set this to TRUE */
+int i_am_filter = FALSE;		     /* After forking, filter will set this to TRUE */
+int ignore_queued_input = FALSE;             /* read and then ignore all characters in input queue until it is empty (i.e. read would block) */
+int received_WINCH = FALSE;                  /* flag set in SIGWINCH signal handler: start line edit as soon as possible */
+int prompt_is_still_uncooked = TRUE;         /* The main loop consults this variable to determine the select() timeout: when TRUE, it is
+                                                a few millisecs, if FALSE, it is infinite.						
+                                                TRUE just after receiving command output (when we still don't know whether we have a
+                                                prompt), and, importantly, at startup (so that substitute prompts get displayed even with
+                                                programs that don't have a startup message, such as cat)  */
 						
-int we_just_got_a_signal_or_EOF = FALSE;  /* When we got a signal or EOF, and the program sends something that ends in a newline, take it
-					     as a response to user input - i.e. preserve a cooked prompt and just print the new output after it */
+int we_just_got_a_signal_or_EOF = FALSE;     /* When we got a signal or EOF, and the program sends something that ends in a newline, take it
+					       as a response to user input - i.e. preserve a cooked prompt and just print the new output after it */
 int rlwrap_already_prompted = FALSE;
-int accepted_lines =  0; /* number of lines accepted (used for one-shot rlwrap) */
-
+int accepted_lines =  0;                     /* number of lines accepted (used for one-shot rlwrap) */
+bool user_has_typed_first_NL = FALSE;        /* When the *first* user NL is typed in direct mode, we probably need --always-readline: give a warning, .. */
+bool advise_always_readline = FALSE;         /* ... which has to be given a bit later in order to not mess up the screen                                 */
 
 /* private variables */
 static char *history_filename = NULL;
@@ -85,33 +90,45 @@ static int  last_opt = -1;
 static char *client_term_name = NULL; /* we'll set TERM to this before exec'ing client command */
 static int feed_history_into_completion_list = FALSE;
 
+
+/*
+ * Since version 0.24, rlwrap only writes to master_pty
+ * asynchronously, keeping a queue of pending output. The readline
+ * line handler calls put_in_output_queue(user_input) , while
+ * main_loop calls flush_output_queue() as long as there is something
+ * in the queue.
+ */
+
+static char *output_queue; /* NULL when empty */
+
+
 /* private functions */
 static void init_rlwrap(char *command_line);
 static void fork_child(char *command_name, char **argv);
 static char *read_options_and_command_name(int argc, char **argv);
 static void main_loop(void);
-static void test_main(void);
 
 
 
 /* options */
 #ifdef GETOPT_GROKS_OPTIONAL_ARGS
-static char optstring[] = "+:a::Ab:cC:d::D:e:f:F:g:hH:iIl:nNM:m::oO:p::P:q:rRs:S:t:TUvw:Wz:";
+static char optstring[] = "+:a::A::b:cC:d::D:e:Ef:F:g:hH:iIl:nNM:m::oO:p::P:q:rRs:S:t:TUvw:Wz:";
 /* +: is not really documented. configure checks wheteher it works as expected
    if not, GETOPT_GROKS_OPTIONAL_ARGS is undefined. @@@ */
 #else
-static char optstring[] = "+:a:Ab:cC:d:D:e:f:F:g:hH:iIl:nNM:m:oO:p:P:q:rRs:S:t:TUvw:Wz:";	
+static char optstring[] = "+:a:A:b:cC:d:D:e:Ef:F:g:hH:iIl:nNM:m:oO:p:P:q:rRs:S:t:TUvw:Wz:";	
 #endif
 
 #ifdef HAVE_GETOPT_LONG
 static struct option longopts[] = {
   {"always-readline", 		optional_argument, 	NULL, 'a'},
-  {"ansi-colour-aware",         no_argument,            NULL, 'A'},
+  {"ansi-colour-aware",         optional_argument,      NULL, 'A'},
   {"break-chars", 		required_argument, 	NULL, 'b'},
   {"complete-filenames", 	no_argument, 		NULL, 'c'},
   {"command-name", 		required_argument, 	NULL, 'C'},
   {"debug", 			optional_argument, 	NULL, 'd'},
   {"extra-char-after-completion", required_argument,    NULL, 'e'},
+  {"always-echo",               no_argument,            NULL, 'E'},
   {"history-no-dupes", 		required_argument, 	NULL, 'D'},
   {"file", 			required_argument, 	NULL, 'f'},
   {"history-format", 		required_argument, 	NULL, 'F'},
@@ -147,6 +164,23 @@ static struct option longopts[] = {
 
 
 
+/* helper function to run a unit test whenever UNIT_TEST is #defined, e.g. by "make clean; make CFLAGS='-DUNIT_TEST=my_test'" */
+
+#ifdef UNIT_TEST
+  static void run_unit_test(int argc, char **argv, test_stage stage) {
+    extern void UNIT_TEST(int argc, char **argv, test_stage stage);
+    #define VALUE_AS_STRING(v) NAME_AS_STRING(v)
+    #define NAME_AS_STRING(n) #n
+  
+    if(stage == TEST_AT_PROGRAM_START)
+      myerror(WARNING|NOERRNO, "running unit test %s", VALUE_AS_STRING(UNIT_TEST)); 
+    UNIT_TEST(argc, argv, stage);
+  }
+#else
+  static void run_unit_test(int UNUSED(argc), char ** UNUSED(argv), test_stage UNUSED(stage)) {
+    /*do nothing */
+  }
+#endif
 /*
  * main function. initialises everything and calls main_loop(),
  * which never returns
@@ -154,27 +188,44 @@ static struct option longopts[] = {
 
 int
 main(int argc, char **argv)
-{ 
+{
   char *command_name;
+
+  if (!(setlocale (LC_ALL, "") && setlocale(LC_COLLATE, "C")))/* ANSI C says that all programs start by default in the standard `C' locale...   */ 
+    myerror(WARNING|NOERRNO, "could not set locale");         /* ... To use the locales specified by the environment, we must call  setlocale.  */
+                                                              /* LC_COLLATE = 'C' because we use character ranges  as byte ranges in regexps    */ 
+  
+  run_unit_test(argc, argv,TEST_AT_PROGRAM_START);
   rlwrap_command_line = unsplit_with(argc, argv, " ");     
   init_completer();
 
   /* Harvest options and leave optind pointing to first non-option argument: */
   command_name = read_options_and_command_name(argc, argv);
 
+
   /* by now, optind points to slave <command>, and &argv[optind] is <command>'s argv. Remember slave command line: */
   command_line = unsplit_with(argc - optind, argv + optind, " ");
+  
+  run_unit_test(argc - optind, argv + optind, TEST_AFTER_OPTION_PARSING); /* argv points at the first non-option rlwrap argument */
 
-  /* if stdin is not a tty, just execute <command>: */ 
-  if (!isatty(STDIN_FILENO) && execvp(argv[optind], &argv[optind]) < 0)
-    myerror(FATAL|USE_ERRNO, "Cannot execute %s", argv[optind]);	
+  /* if stdin is not a tty, or we're inside emacs, just execute <command>: */ 
+  if (!isatty(STDIN_FILENO) || getenv("INSIDE_EMACS"))
+    if (execvp(argv[optind], &argv[optind]) < 0)
+      myerror(FATAL|USE_ERRNO, "Cannot execute %s", argv[optind]);
+  
   init_rlwrap(rlwrap_command_line);
   install_signal_handlers();	
   block_all_signals();
   fork_child(command_name, argv); /* this will unblock most signals most of the time */
+
+  init_readline("");
+  last_minute_checks();
+  run_unit_test(0,NULL, TEST_AFTER_READLINE_INIT); 
+
   if (filter_command)
     spawn_filter(filter_command);
-  
+  run_unit_test(argc - optind, argv + optind, TEST_AFTER_SPAWNING_SLAVE_COMMAND); /* argv points at the first non-option rlwrap argument */
+
   main_loop();
   return 42;			/* The Answer, but, sadly, we'll never get there.... */
 }
@@ -236,7 +287,6 @@ main_loop()
   int leave_prompt_alone;
   sigset_t no_signals_blocked;
   int seen_EOF = FALSE;     
-
    
   struct timespec         select_timeout, *select_timeoutptr;
   struct timespec immediately = { 0, 0 }; /* zero timeout when child is dead */
@@ -248,12 +298,9 @@ main_loop()
   sigemptyset(&no_signals_blocked);
   
 
-  init_readline("");
-  last_minute_checks();
+
   pass_through_filter(TAG_OUTPUT,""); /* If something is wrong with filter, get the error NOW */
-  set_echo(FALSE);		/* This will also put the terminal in CBREAK mode */
-  test_main(); 
-  
+  set_echo(FALSE);		      /* This will also put the terminal in CBREAK mode */
   /* ------------------------------  main loop  -------------------------------*/
   while (TRUE) {
     /* listen on both stdin and pty_fd */
@@ -282,14 +329,22 @@ main_loop()
       timeoutstr = "forever";
     }
      
-    DPRINTF1(DEBUG_TERMIO, "calling select() with timeout %s",  timeoutstr);
+    DPRINTF2(DEBUG_TERMIO, "calling select() with timeout %s %s ...",  timeoutstr, within_line_edit ? "(within line edit)" : "");
     
 
     nfds = my_pselect(1 + master_pty_fd, &readfds, &writefds, NULL, select_timeoutptr, &no_signals_blocked);
 
-    DPRINTF3(DEBUG_TERMIO, "select() returned  %d (stdin|pty in|pty out = %03d), within_line_edit=%d", nfds,
-	     100*(FD_ISSET(STDIN_FILENO, &readfds)?1:0) + 10*(FD_ISSET(master_pty_fd, &readfds)?1:0) + (FD_ISSET(master_pty_fd, &writefds)?1:0), 
-	     within_line_edit);
+    /* DPRINTF3(DEBUG_TERMIO, "select() returned  %d (stdin|pty in|pty out = %03d), within_line_edit=%d", nfds, */
+    /*          100*(FD_ISSET(STDIN_FILENO, &readfds)?1:0) + 10*(FD_ISSET(master_pty_fd, &readfds)?1:0) + (FD_ISSET(master_pty_fd, &writefds)?1:0),  */
+    /*          within_line_edit); */
+
+    
+    DPRINTF5(DEBUG_TERMIO, "... returning %d%s %s %s %s"
+             , nfds
+             , nfds > 0 ? ": " : "."
+	     , nfds > 0 && FD_ISSET(STDIN_FILENO, &readfds)   ? "stdin ready for input" : ""
+             , nfds > 0 && FD_ISSET(master_pty_fd, &readfds)  ? "pty master ready for input": ""
+             , nfds > 0 && FD_ISSET(master_pty_fd, &writefds) ? "output queue nonempty and pty master ready for output" : "");
 
     assert(!filter_pid || filter_is_dead || kill(filter_pid,0) == 0); 
     assert(command_is_dead || kill(command_pid,0) == 0);
@@ -344,9 +399,7 @@ main_loop()
 	  /* move and erase before cooking, as we need to move/erase according
 	     to the raw prompt */
           cook_prompt_if_necessary();
-	  DPRINTF2(DEBUG_READLINE,"After cooking, raw_prompt=%s, cooked=%s",
-                   mangle_string_for_debug_log(saved_rl_state.raw_prompt, MANGLE_LENGTH),
-                   mangle_string_for_debug_log(saved_rl_state.cooked_prompt, MANGLE_LENGTH));
+	  DPRINTF2(DEBUG_READLINE,"After cooking, raw_prompt=%s, cooked=%s", M(saved_rl_state.raw_prompt), M(saved_rl_state.cooked_prompt));
 	  my_putstr(saved_rl_state.cooked_prompt);
 	  rlwrap_already_prompted = TRUE;
 	}
@@ -396,25 +449,32 @@ main_loop()
             continue;                   /* and try one more time (hopefully catching the signal this time round                 */
           } else {
             myerror(FATAL|USE_ERRNO, "read error on master pty"); 
-          }
+          } 
 	}
-	  
+        remove_padding_and_terminate(buf, nread);
 	completely_mirror_slaves_output_settings(); /* some programs (e.g. joe) need this. Gasp!! */	
         mirror_args(command_pid);        
-	
+	check_cupcodes(buf);
         if (skip_rlwrap()) { /* Race condition here! The client may just have finished an emacs session and
 			        returned to cooked mode, while its ncurses-riddled output is stil waiting for us to be processed. */
-	  write_patiently(STDOUT_FILENO, buf, nread, "to stdout");
-
-	  DPRINTF2(DEBUG_TERMIO, "read from pty and wrote to stdout  %d  bytes in direct mode  <%s>",
-                   nread, mangle_string_for_debug_log((buf[nread]='\0', buf), MANGLE_LENGTH));
+          if (advise_always_readline) {
+            char *newlines = "\n\n";
+            assert(!always_readline);
+            write_patiently(STDOUT_FILENO, newlines, strlen(newlines), "to stdout"); /* make the following warning stand out ...  */
+            myerror(WARNING|NOERRNO, "rlwrap appears to do nothing for %s, which asks for\n"
+                                     "single keypresses all the time. Don't you need --always-readline\n"
+                                     "and possibly --no-children? (cf. the rlwrap manpage)\n", command_name);
+            advise_always_readline = FALSE;
+          }
+          write_patiently(STDOUT_FILENO, buf, nread, "to stdout"); /* ... and print it before the clients output */
+          DPRINTF1(DEBUG_AD_HOC, "advise_always_readline = %d", advise_always_readline);
+	  DPRINTF2(DEBUG_TERMIO, "read from pty and wrote to stdout  %d  bytes in direct mode  <%s>",  nread, M(buf));
 	  yield();
 	  continue;
 	}
 
-	DPRINTF2(DEBUG_TERMIO, "read %d bytes from pty into buffer: %s", nread,  mangle_string_for_debug_log((buf[nread]='\0', buf), MANGLE_LENGTH));
+	DPRINTF2(DEBUG_TERMIO, "read %d bytes from pty into buffer: %s", nread, M(buf));
         
-        remove_padding_and_terminate(buf, nread);
         
 	write_logfile(buf);
 	if (within_line_edit)	/* client output arrives while we're editing keyboard input:  */
@@ -445,7 +505,7 @@ main_loop()
                                                                      response to user input, so leave the prompt alone */
 
         DPRINTF3(DEBUG_READLINE, "leave_prompt_alone: %s (raw prompt: %s, prompt_is_still_uncooked: %d)",
-                 (leave_prompt_alone? "yes" : "no"), mangle_string_for_debug_log(saved_rl_state.raw_prompt, MANGLE_LENGTH), prompt_is_still_uncooked);
+                 (leave_prompt_alone? "yes" : "no"), M(saved_rl_state.raw_prompt), prompt_is_still_uncooked);
 	
         if (!leave_prompt_alone) /* && (!impatient_prompt || !saved_rl_state.cooked_prompt)) */
 	  move_cursor_to_start_of_prompt(ERASE);  
@@ -467,8 +527,16 @@ main_loop()
             my_putstr(old_raw_prompt);
             free(old_raw_prompt);
           }
-	  my_putstr(filtered);
+
+	  my_putstr(filtered); 
 	  free (filtered);
+          if (regexp_means_prompt && prompt_regexp && match_regexp(saved_rl_state.raw_prompt, prompt_regexp, FALSE)) {
+            /* user specified -O!.... so any natching candidate prompt will be cooked and output immediately: */
+            move_cursor_to_start_of_prompt(ERASE);  /* erase already printed raw prompt */
+            cook_prompt_if_necessary();             
+            my_putstr(saved_rl_state.cooked_prompt);
+          }     
+            
 	  rlwrap_already_prompted = TRUE;
 	} else {
 	  my_putstr(new_output_minus_prompt);
@@ -516,7 +584,12 @@ main_loop()
 	if (skip_rlwrap()) {	/* direct mode, just pass it on */
 	                        /* remote possibility of a race condition here: when the first half of a multi-byte char is read in
 				   direct mode and the second half in readline mode. Oh well... */
-	  DPRINTF0(DEBUG_TERMIO, "passing it on (in transparent mode)");	
+	  DPRINTF0(DEBUG_TERMIO, "passing it on (in transparent mode)");
+          if (!user_has_typed_first_NL && (byte_read == '\r' || byte_read == '\n')) {
+            user_has_typed_first_NL = TRUE;
+            advise_always_readline = TRUE; /* first NL is in direct mode: advise the user that she probably wants --always-readline */ 
+          }
+            
 	  completely_mirror_slaves_terminal_settings(); /* this is of course 1 keypress too late: we should
 							   mirror the terminal settings *before* the user presses a key.
 							   (maybe using rl_event_hook??)   @@@FIXME  @@@ HOW?*/
@@ -548,9 +621,11 @@ main_loop()
       /* -------------------------- write pty --------------------------------- */
       if (FD_ISSET(master_pty_fd, &writefds)) {
 	flush_output_queue();
-	yield(); /*  give  slave command time to respond. If we don't do this,
-		     nothing bad will happen, but the "dialogue" on screen will be
-		     out of order   */
+        if(output_queue) {   /* there was more than one line in the queue - probably pasted input    */
+          mymicrosleep(10);  /* give slave some time to respond                                      */
+	  yield();           /*  If we woudn't do this, nothing bad would happen, but the            */
+                             /*  "dialogue" on screen will be out of order (which can still happen)  */
+        }
       }
     }				/* if (ndfs > 0)         */
   }				/* while (1)             */
@@ -579,9 +654,13 @@ init_rlwrap(char *command_line)
   
   /* Determine rlwrap home dir and prefix for default history and completion filenames */
   homedir = (getenv("RLWRAP_HOME") ? getenv("RLWRAP_HOME") : getenv("HOME"));
+  if (!homedir) {
+    homedir = ".";
+    myerror(WARNING | NOERRNO, "No HOME, using '%s'", homedir);
+  }
   homedir_prefix = (getenv("RLWRAP_HOME") ?                    /* is RLWRAP_HOME set?                */
 		    add2strings(getenv("RLWRAP_HOME"), "/") :  /* use $RLWRAP_HOME/<command>_history */
-		    add2strings(getenv("HOME"), "/."));	       /* if not, use ~/.<command>_history   */
+		    add2strings(homedir, "/."));	       /* if not, use ~/.<command>_history   */
 
   /* Determine history file name and check its existence and permissions */
 
@@ -636,6 +715,7 @@ init_rlwrap(char *command_line)
     add3strings(DATADIR, "/rlwrap/completions/", command_name);
 
   rl_readline_name = command_name;
+  DPRINTF1(DEBUG_READLINE, "Setting readline command name to '%s'", command_name);
 
   /* Initialise completion list (if <completion_filename> is readable) */
   if (access(completion_filename, R_OK) == 0) {
@@ -654,9 +734,10 @@ init_rlwrap(char *command_line)
  */
 
 static char *
-check_optarg(char opt, int remaining)
+check_optarg(char opt, int remaining, bool isdummy, const char *suggestion)
 {
-  MAYBE_UNUSED2(opt, remaining);
+
+  MAYBE_UNUSED(opt); MAYBE_UNUSED(remaining); MAYBE_UNUSED(isdummy); MAYBE_UNUSED(suggestion);
   
   if (!optarg)
     last_option_didnt_have_optional_argument = TRUE; /* if this variable is set, and  if command is not found,
@@ -664,13 +745,13 @@ check_optarg(char opt, int remaining)
 						        as optional argument (e.g. 'rlwrap -a password sqlplus' will try to
 						        execute 'password sqlplus' ) */
 #ifndef GETOPT_GROKS_OPTIONAL_ARGS
-  if (optarg &&    /* is there an optional arg? have a look at it: */
-      ((optarg[0] == '-' && isalpha(optarg[1])) || /* looks like next option */
+  if (optarg &&         /* is there an optional arg? have a look at it: */
+      ((optarg[0] == '-' && (optarg[1] == '-' || isalpha(optarg[1]))) || /* looks like next option */
        remaining == 0)) /* or is last item on command line */
 
     myerror(WARNING|NOERRNO, "on this system, the getopt() library function doesn't\n"
        "grok optional arguments, so '%s' is taken as an argument to the -%c option\n"
-       "Is this what you meant? If not, please provide an argument", optarg, opt);
+            "Is this what you meant? If not, please provide %s  argument like '%s'", optarg, opt, isdummy ? "a dummy" : "an", suggestion);
 #endif
   
   return optarg;
@@ -729,10 +810,14 @@ read_options_and_command_name(int argc, char **argv)
     switch (c) {
     case 'a':
       always_readline = TRUE;
-      if (check_optarg('a', remaining))
+      if (check_optarg('a', remaining, TRUE, "brumbleslurgh"))
         password_prompt_search_string = mysavestring(optarg);
       break;
-    case 'A':	ansi_colour_aware = TRUE; break;
+    case 'A':
+      ansi_colour_aware = TRUE;
+      if (check_optarg('A', remaining, TRUE, "") && strings_are_equal(optarg, "!"))
+          bleach_the_prompt = TRUE;
+      break;
     case 'b':
       rl_basic_word_break_characters = add3strings("\r\n \t", optarg, "");
       opt_b = TRUE;
@@ -749,11 +834,8 @@ read_options_and_command_name(int argc, char **argv)
       if (option_count > 1)
         myerror(FATAL|NOERRNO, "-d or --debug option has to be the *first* rlwrap option\n"
                                "in order to be able to follow the processing  of all subsequent options");
-      debug = check_optarg('d', remaining) ? my_atoi(optarg) : DEBUG_DEFAULT;
-      debug_fp = fopen(DEBUG_FILENAME, "w");
-      if (!debug_fp)
-         myerror(FATAL|USE_ERRNO, "Couldn't open debug file %s", DEBUG_FILENAME);
-      setbuf(debug_fp, NULL); /* always write debug messages to disk at once */
+      debug = check_optarg('d', remaining, FALSE, "7") ? my_atoi(optarg) : DEBUG_DEFAULT;
+      my_fopen(&debug_fp, DEBUG_FILENAME, "w+", "debug log"); /* w+, not w: both parent and child write to the same logfile, and need to fseek beyond where the other may have written stuff */
 #else
       myerror(FATAL|NOERRNO, "To use -d( for debugging), configure %s with --enable-debug and rebuild",program_name);
 #endif
@@ -770,6 +852,7 @@ read_options_and_command_name(int argc, char **argv)
       if (strlen(extra_char_after_completion) > 1) 
         myerror(FATAL|NOERRNO, "-e (--extra-char-after-completion) argument should be at most one character");
       break;
+    case 'E': always_echo = TRUE; break;
     case 'f':
       if (strcmp(optarg, ".") == 0)
         feed_history_into_completion_list =  TRUE;
@@ -777,9 +860,9 @@ read_options_and_command_name(int argc, char **argv)
         feed_file_into_completion_list(optarg);
       opt_f = TRUE;
       break;
-    case 'F': myerror(FATAL|NOERRNO, "The -F (--history-format) option is obsolete. Use -z \"history_format '%s'\" instead", optarg);
-    case 'g': forget_regexp = mysavestring(optarg); match_regexp("just testing", forget_regexp, 1); break;
-    case 'h': usage(EXIT_SUCCESS);		/* will call exit() */
+    case 'F': WONTRETURN(myerror(FATAL|NOERRNO, "The -F (--history-format) option is obsolete. Use -z \"history_format '%s'\" instead", optarg));
+    case 'g': forget_regexp = mysavestring(optarg);  match_regexp("complain NOW if regexp is wrong", forget_regexp, 1); break;
+    case 'h': WONTRETURN(usage(EXIT_SUCCESS));	  
     case 'H': history_filename = mysavestring(optarg); break;
     case 'i': 
       if (opt_f)
@@ -793,17 +876,23 @@ read_options_and_command_name(int argc, char **argv)
 #ifndef HAVE_SYSTEM
       myerror(WARNING|NOERRNO, "the -m option doesn't work on this system");
 #endif
-      multiline_separator =
-        (check_optarg('m', remaining) ? mysavestring(optarg) : " \\ ");
+      multiline_separator = /* \\\\ will be printed as \\ which is correct if we want ' \ ' to be the multiline separator */
+        (check_optarg('m', remaining, FALSE, " \\\\ ") ? mysavestring(optarg) : " \\ ");
       break;
     case 'M': multi_line_tmpfile_ext = mysavestring(optarg); break;
     case 'N': commands_children_not_wrapped = TRUE; break;
     case 'o': 
       one_shot_rlwrap = TRUE; break;
-    case 'O': prompt_regexp = mysavestring(optarg); match_regexp("just testing", prompt_regexp, 1); break;
+    case 'O': prompt_regexp = mysavestring(optarg);
+      if (*prompt_regexp == '!') {
+        regexp_means_prompt = TRUE;
+        prompt_regexp += 1;
+      }
+      match_regexp("complain NOW if regexp is wrong", prompt_regexp, 1);
+      break;
     case 'p':
       colour_the_prompt = TRUE;
-      initialise_colour_codes(check_optarg('p', remaining) ?
+      initialise_colour_codes(check_optarg('p', remaining, FALSE, "Red") ?
                               colour_name_to_ansi_code(optarg) :
                               colour_name_to_ansi_code("Red"));
       break;
@@ -841,11 +930,10 @@ read_options_and_command_name(int argc, char **argv)
     case 'z': filter_command = mysavestring(optarg);	break;
     case '?':
       assert(optind > 0);
-      myerror(FATAL|NOERRNO, "unrecognised option %s\ntry '%s --help' for more information", argv[optind-1], full_program_name);
+      WONTRETURN(myerror(FATAL|NOERRNO, "unrecognised option %s\ntry '%s --help' for more information", argv[optind-1], full_program_name));
     case ':':
       assert(optind > 0);
-      myerror(FATAL|NOERRNO, "option %s requires an argument \ntry '%s --help' for more information",
-         argv[optind-1], full_program_name);
+      WONTRETURN(myerror(FATAL|NOERRNO, "option %s requires an argument \ntry '%s --help' for more information", argv[optind-1], full_program_name));
 
     default:
       usage(EXIT_FAILURE);
@@ -860,7 +948,7 @@ read_options_and_command_name(int argc, char **argv)
   
   if (optind >= argc) { /* rlwrap -a -b -c with no command specified */
     if (filter_command) { /* rlwrap -z filter with no command specified */
-      mysignal(SIGALRM, &handle_sigALRM); /* needed for read_patiently2 */
+      mysignal(SIGALRM, HANDLER(handle_sigALRM)); /* needed for read_patiently2 */
       spawn_filter(filter_command);
       pass_through_filter(TAG_OUTPUT,""); /* ignore result but allow TAG_OUTPUT_OUT_OF_BAND */
       cleanup_rlwrap_and_exit(EXIT_SUCCESS);
@@ -904,16 +992,6 @@ read_options_and_command_name(int argc, char **argv)
 }
 
 
-/*
- * Since version 0.24, rlwrap only writes to master_pty
- * asynchronously, keeping a queue of pending output. The readline
- * line handler calls put_in_output_queue(user_input) , while
- * main_loop calls flush_output_queue() as long as there is something
- * in the queue.
- */
-
-static char *output_queue; /* NULL when empty */
-
 int
 output_queue_is_nonempty()
 {
@@ -925,8 +1003,7 @@ put_in_output_queue(char *stuff)
 {
   
   output_queue = append_and_free_old(output_queue, stuff); 
-  DPRINTF3(DEBUG_TERMIO,"put %d bytes in output queue (which now has %d bytes): %s",
-	   (int) strlen(stuff), (int) strlen(output_queue), mangle_string_for_debug_log(stuff, 20));
+  DPRINTF3(DEBUG_TERMIO,"put %d bytes in output queue (which now has %d bytes): %s", (int) strlen(stuff), (int) strlen(output_queue), M(stuff));
 
 }
 
@@ -954,8 +1031,7 @@ flush_output_queue()
   if (debug) {
     char scratch = output_queue[nwritten];
     output_queue[nwritten] = '\0'; /* temporarily replace the last written byte + 1 by a '\0' */
-    DPRINTF3(DEBUG_TERMIO,"flushed %d of %d bytes from output queue to pty: %s",
-	     nwritten, queuelen, mangle_string_for_debug_log(output_queue, MANGLE_LENGTH));
+    DPRINTF3(DEBUG_TERMIO,"flushed %d of %d bytes from output queue to pty: %s", nwritten, queuelen, M(output_queue));
     output_queue[nwritten] =  scratch;
   }	
   
@@ -988,7 +1064,7 @@ cleanup_rlwrap_and_exit(int status)
     write_history(history_filename);	/* ignore errors */
 
   close_logfile();
-
+  
   DPRINTF4(DEBUG_SIGNALS, "command_pid: %d, commands_exit_status: %x, filter_pid: %d, filters_exit_status: %x",
            command_pid, commands_exit_status, filter_pid, filters_exit_status);
   mymicrosleep(10); /* we may have got an EOF or EPIPE because the filter or command died, but this doesn't mean that
@@ -1003,7 +1079,12 @@ cleanup_rlwrap_and_exit(int status)
   }     
   if (debug) 
     debug_postmortem();
-  
+
+
+  if (bracketed_paste_enabled) {
+    DPRINTF0(DEBUG_READLINE, "disabling bracketed-paste");
+    my_putstr(term_disable_bracketed_paste);
+  }
   if (terminal_settings_saved)
     if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_terminal_settings) < 0)  /* ignore errors (almost dead anyway) */ 
       { /* nothing ... */ } /* fprintf(stderr, "Arggh\n"); don't use myerror!!*/
@@ -1025,25 +1106,3 @@ cleanup_rlwrap_and_exit(int status)
 }
 
 
-
-extern int test_haystack(const char *haystack, const char *needle);
-static void test_main() {
-#ifdef DEBUG
-  if(debug & DEBUG_TEST_MAIN) {
-    /* test, test */
-    test_haystack("rabarber","r");
-    test_haystack("","r");
-    test_haystack("rr rr rr rrr","rr");
-    /* test_haystack("bobobo",""); */
-    test_haystack("lalalalalal","la");
-
-    test_unbackspace("abc\rd","dbc");
-    test_unbackspace("abc\bd\re","ebd");
-    test_unbackspace("abc\rde\rf","fec");
-    puts("tests OK");
-    exit(0);
-  }     
-      
-#endif      
-}	
-
