@@ -19,6 +19,7 @@
     e-mail:  hanslub42@gmail.com
 */
 
+
 #include "../config.h"
 #include <sys/types.h>
 #if HAVE_SYS_WAIT_H
@@ -50,11 +51,14 @@
    #define __USE_GNU */
 
 #include <stdlib.h>
+#include <locale.h>
+
 
 #include <sched.h>
 
 
 
+
 #if TIME_WITH_SYS_TIME
 # include <sys/time.h>
 # include <time.h>
@@ -66,11 +70,19 @@
 # endif
 #endif
 
+#if HAVE_SYS_FILE_H
+#include <sys/file.h>
+#endif
 
 #ifdef HAVE_GETOPT_H
 #  include <getopt.h>
 #endif
 
+#ifdef HAVE_STROPTS_H
+#  include <stropts.h>
+#endif
+
+
 
 #ifdef HAVE_LIBGEN_H
 #  include <libgen.h>
@@ -137,7 +149,7 @@ char *strchr(), *strrchr();
 #  define CAN_FOLLOW_COMMANDS_CWD 1
 #endif
 
-#define BUFFSIZE 512
+#define BUFFSIZE 2048
 
 #ifndef MAXPATHLEN
 #define MAXPATHLEN 512
@@ -190,7 +202,7 @@ extern int _rl_horizontal_scroll_mode;
 #  define MAYBE_MULTILINE 1
 # endif
 #else
-# define _rl_eof_char 0;
+# define _rl_eof_char 0
 #endif
 
 #ifdef MAYBE_MULTILINE
@@ -204,18 +216,26 @@ extern int _rl_horizontal_scroll_mode;
 
 
 
+extern void test_main(void);
+
 /* in main.c: */
 extern int master_pty_fd;
 extern int slave_pty_sensing_fd;
 extern FILE *debug_fp;
 extern char *program_name, *command_name;
 extern int always_readline;
+extern int always_echo;
 extern int complete_filenames;
+extern int within_line_edit;
+extern int screen_is_alternate;
 extern pid_t command_pid;
 extern char *rlwrap_command_line;
+extern int unit_test_argc;
+extern char **unit_test_argv;
 extern char *command_line;
 extern char *extra_char_after_completion;
 extern int i_am_child;
+extern int i_am_filter;
 extern int nowarn;
 extern int debug;
 extern char *password_prompt_search_string;
@@ -235,6 +255,7 @@ extern int pass_on_sigINT_as_sigTERM;
 #define ELIMINATE_ALL_DOUBLES           2
 extern int one_shot_rlwrap;
 extern int ansi_colour_aware;
+extern int bleach_the_prompt;
 extern int colour_the_prompt;
 extern int received_WINCH;
 extern int prompt_is_still_uncooked;
@@ -277,11 +298,11 @@ int prompt_is_single_line(void);
 char *process_new_output(const char* buffer, struct rl_state* state);
 int cook_prompt_if_necessary (void);
 
-extern int within_line_edit, transparent; 
+extern int transparent; 
 extern char *multiline_separator;
 extern char *pre_given;
 extern int leave_prompt_alone;
-
+extern bool bracketed_paste_enabled;
 
 
 /* in signals.c */
@@ -299,8 +320,10 @@ extern int received_sigALRM;
 #endif
 typedef RETSIGTYPE (*sighandler_type)(int);       
 
+/* we'll install signal handlers with mysignal(SIGBLAH, HANDLER(handler)), to improve debug log readabilty */
+void mysignal(int sig, sighandler_type handler, const char *handler_name);
+#define HANDLER(f) &f, #f
 
-void mysignal(int sig, sighandler_type handler);
 void install_signal_handlers(void);
 void block_signals(int *sigs);
 void unblock_signals(int *sigs);
@@ -346,6 +369,7 @@ void  free_multiple(void *ptr, ...);
 void  mysetsid(void);
 void  close_open_files_without_writing_buffers(void);
 size_t filesize(const char *filename);
+void  my_fopen(FILE  **pfp, const char *path, const char *mode, const char *description);
 void  open_logfile(const char *filename);
 void  write_logfile(const char *str);
 void  close_logfile(void);
@@ -365,8 +389,10 @@ char *mybasename(const char *filename);
 char *mydirname(const char *filename);
 void  mystrlcpy(char *dst, const char *src, size_t size);
 void  mystrlcat(char *dst, const char *src, size_t size);
+bool strings_are_equal(const char *s1, const char *s2);
 char *mystrstr(const char *haystack, const char *needle);
 char *mysavestring(const char *string);
+char *strifnull(char *string);
 char *add3strings(const char *str1, const char *str2, const char *str3);
 #define add2strings(a,b)  add3strings(a,b,"")
 int my_atoi(const char *nptr);
@@ -389,7 +415,6 @@ char *as_string(int i);
 char *append_and_expand_history_format(char *line);
 void remove_padding_and_terminate(char *buf, int length);
 void unbackspace(char* buf);
-void test_unbackspace (const char *input, const char *expected_result);
 char *mark_invisible(const char *buf);
 char *copy_and_unbackspace(const char *original);
 int colourless_strlen(const char *str, char **pcopy_without_ignore_markers, int termwidth);
@@ -403,6 +428,8 @@ int isnumeric(char *string);
 char *append_field_and_free_old(char *message, const char *field);
 char *merge_fields(char *field, ...);
 char **split_filter_message(char *message, int *count);
+char *protect_or_cleanup(char *prompt, bool free_prompt);
+void check_cupcodes(const char *client_output);
 
 /* in pty.c: */
 pid_t my_pty_fork(int *, const struct termios *, const struct winsize *);
@@ -455,7 +482,9 @@ extern int terminal_settings_saved;
 extern struct winsize winsize;
 extern char *term_name;
 extern char *term_backspace, term_eof, term_stop, *term_cursor_hpos,
-  *term_cursor_up, *term_cursor_down, *term_cursor_left, *term_rmcup, *term_rmkx;
+  *term_cursor_up, *term_cursor_down, *term_cursor_left, *term_smcup,
+  *term_rmcup, *term_rmkx, *term_enable_bracketed_paste, *term_disable_bracketed_paste;
+extern int term_has_colours;
 extern int newline_came_last;
 
 /* in filter.c */
@@ -467,7 +496,8 @@ extern int newline_came_last;
 #define TAG_COMPLETION 3
 #define TAG_PROMPT 4
 #define TAG_HOTKEY 5
-#define MAX_INTERESTING_TAG 5 /* max tag for which the filter can have a handler */
+#define TAG_SIGNAL 6
+#define MAX_INTERESTING_TAG 6 /* max tag for which the filter can have a handler */
 
 #define TAG_WHAT_ARE_YOUR_INTERESTS 127
 
@@ -485,12 +515,34 @@ extern pid_t filter_pid;
 extern int filter_is_dead;
 void spawn_filter(const char *filter_command);
 void kill_filter(void);
+int filter_is_interested_in(int tag); 
 char *pass_through_filter(int tag, const char *buffer);
 char *filters_last_words(void);
 void filter_test(void);
 
 
 
+/* in multibyte.c: */
+
+#ifdef MULTIBYTE_AWARE /* i.e. if configured with --enable-multibyte-aware  */
+   #include <wchar.h>
+   #define MBSTATE mbstate_t
+#else
+   #define MBSTATE int
+#endif
+
+MBSTATE * mbc_initstate(MBSTATE *st);
+MBSTATE *mbc_copystate(MBSTATE st, MBSTATE *stc);
+int mbc_is_valid(const char *mb_string, const MBSTATE *st);
+const char * mbc_next(const char *mb_string, MBSTATE *st);
+int mbc_charwidth(const char *p, MBSTATE *st);
+char *mbc_first(const char *mb_string, const MBSTATE *st);
+int is_multibyte(const char *mb_char, const MBSTATE *st);
+const char *mbc_inc(const char **mbc, MBSTATE *st);
+void mbc_copy(const char *p, char **q, MBSTATE *st);
+size_t mbc_strnlen(const char *mb_string, size_t maxlen, MBSTATE *st);
+
+
 
 /* some handy macros */
 
@@ -502,6 +554,7 @@ void filter_test(void);
 #  define FALSE 0
 #endif
 
+
 #ifndef min
 # define min(a,b) ((a) < (b) ? (a) : (b))
 #endif
@@ -511,6 +564,11 @@ void filter_test(void);
 #endif
 
 
+/* macros for writing unit tests */
+typedef enum  {TEST_AT_PROGRAM_START, TEST_AFTER_OPTION_PARSING, TEST_AFTER_SPAWNING_SLAVE_COMMAND,TEST_AFTER_READLINE_INIT} test_stage;
+#define TESTFUNC(f, argc, argv, stage)   void f(int argc, char **argv, test_stage stage)
+#define ONLY_AT_STAGE(s) if(stage != s) return
+
 
 
 #include "malloc_debug.h" /* malloc_debug.{c,h} not ready for prime time */
@@ -522,6 +580,23 @@ void filter_test(void);
 
 /* DPRINTF0 and its ilk  doesn't produce any output except when DEBUG is #defined (via --enable-debug configure option) */
 
+/* Use a superfluous "break" (withn a switch) to prevent "this statement may fall through" warnings when we know that a statement will end the program anyway */
+#define WONTRETURN(statement) statement;break
+
+
+# ifdef __GNUC__
+#   define __MYFUNCTION__ __extension__ __FUNCTION__
+#   define UNUSED(x) UNUSED_ ## x __attribute__((__unused__))
+#   define UNUSED_FUNCTION(f) __attribute__((__unused__)) UNUSED_ ## f
+# else
+#   define __MYFUNCTION__ ""
+#   define UNUSED(x) UNUSED_ ## x
+#   define UNUSED_FUNCTION(f) UNUSED_ ## f  
+# endif
+
+
+#define MAYBE_UNUSED(x) (void) (x)
+
 #ifdef  DEBUG
 
 
@@ -537,71 +612,71 @@ void filter_test(void);
 #  define DEBUG_LONG_STRINGS                     512 /* log all strings completely, however long they are */ 
 #  define DEBUG_RACES                            1024 /* introduce random delays */  
 #  define DEBUG_RANDOM_FAIL                      2048 /* fail tests randomly */
-#  define DEBUG_TEST_MAIN                        4096 /* run test_main and exit  */
+#  define DEBUG_SHOWCURSOR                       4096 /* run test_main and exit  */
 
-#  define DEBUG_MAX                              DEBUG_TEST_MAIN
+#  define DEBUG_MAX                              DEBUG_SHOWCURSOR
 #  define MANGLE_LENGTH                          ((debug_saved & DEBUG_LONG_STRINGS) ? 0 : 20) /* debug_saved is defined within DPRINTF macro */ 
 #  define DEBUG_DEFAULT                          (DEBUG_TERMIO | DEBUG_SIGNALS | DEBUG_READLINE)
 #  define DEBUG_ALL                              (2*DEBUG_MAX-1)
 
-# define MAYBE_UNUSED(x)                         (void) (x)
-# define MAYBE_UNUSED2(x,y)                      (void) (x); (void) (y)
+#  define M(x)                                   mangle_string_for_debug_log(x, MANGLE_LENGTH) 
+
+
+#  define ONLY_USED_FOR_DEBUGGING(x) x
+
+
+#  ifndef HAVE_FLOCK
+#    define flock(x,y)
+#  endif
 
-# ifdef __GNUC__
-#   define __MYFUNCTION__ __extension__ __FUNCTION__
-#   define UNUSED(x) UNUSED_ ## x __attribute__((__unused__))
-#   define UNUSED_FUNCTION(f) __attribute__((__unused__)) UNUSED_ ## f
-# else
-#   define __MYFUNCTION__ ""
-#   define UNUSED(x) UNUSED_ ## x
-#   define UNUSED_FUNCTION(f) UNUSED_ ## f  
-# endif
 
 # define WHERE_AND_WHEN \
-  int debug_saved = debug; char file_line[100], when[100];              \
+  int debug_saved = debug; char file_line[100], when[100];                          \
   if(debug & DEBUG_WITH_TIMESTAMPS) timestamp(when, sizeof(when)); else *when='\0'; \
-  debug = 0; /* don't debug while evaluating the DPRINTF arguments */ \
-  snprintf2(file_line, sizeof(file_line),"%.15s:%d:",__FILE__,__LINE__); \
-  fprintf(debug_fp, "%-20s %s %-25.25s ", file_line, when, __MYFUNCTION__);\
+  debug = 0; /* don't debug while evaluating the DPRINTF arguments */               \
+  snprintf2(file_line, sizeof(file_line), "%.15s:%d:", __FILE__, __LINE__);         \
+  flock(fileno(debug_fp), LOCK_EX);                                                 \
+  fseek(debug_fp, 0, SEEK_END);                                                     \
+  fprintf(debug_fp, "%-20s %s %-7s %-25.25s ", file_line, when, (i_am_child ? "child" : i_am_filter? "filter" :"parent"), __MYFUNCTION__);
 
 
-#  define NL_AND_FLUSH           fputc('\n', debug_fp) ; fflush(debug_fp); debug = debug_saved;
+#  define NL_AND_FLUSH           fputc('\n', debug_fp) ; fflush(debug_fp); flock(fileno(debug_fp), LOCK_UN); debug = debug_saved
 
 #  define DPRINTF0(mask, format)                                        \
-  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format); NL_AND_FLUSH; }
+  do {if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format); NL_AND_FLUSH; }} while(FALSE)
 
 #  define DPRINTF1(mask, format,arg)                                    \
-  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg); NL_AND_FLUSH; }
+  do {if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg); NL_AND_FLUSH; }} while(FALSE)
 
 #  define DPRINTF2(mask, format,arg1, arg2)                             \
-  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2); NL_AND_FLUSH; }
+  do {if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2); NL_AND_FLUSH; }} while (FALSE)
 
 #  define DPRINTF3(mask, format,arg1, arg2, arg3)                       \
-  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2, arg3); NL_AND_FLUSH; }
+  do {if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2, arg3); NL_AND_FLUSH; }} while (FALSE)
 
 #  define DPRINTF4(mask, format,arg1, arg2, arg3, arg4)                 \
-  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2, arg3,arg4); NL_AND_FLUSH; }
+  do {if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2, arg3, arg4); NL_AND_FLUSH; }} while (FALSE)
+
+#  define DPRINTF5(mask, format,arg1, arg2, arg3, arg4, arg5)                \
+  do {if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2, arg3, arg4, arg5); NL_AND_FLUSH; }} while (FALSE)
+
 
 #  define ERRMSG(b)              (b && (errno != 0) ? add3strings("(", strerror(errno), ")") : "" )
 
-#  define SHOWCURSOR(c)          if (debug & DEBUG_READLINE) {my_putchar(c); mymicrosleep(1200); curs_left();} /* (may work incorrectly at last column!)*/
+#  define SHOWCURSOR(c)          if (debug & DEBUG_SHOWCURSOR) {my_putchar(c); mymicrosleep(1000); curs_left();} /* (may work incorrectly at last column!)*/
 
 #  define DEBUG_RANDOM_SLEEP        if (debug & DEBUG_RACES) {int sleeptime=rand()&31; DPRINTF1(DEBUG_RACES,"sleeping for %d msecs", sleeptime); mymicrosleep(sleeptime);}
 
 
 #else
-#  define UNUSED(x) x
-#  define UNUSED_FUNCTION(f) f
-#  define MAYBE_UNUSED(x)
-#  define MAYBE_UNUSED2(x,y)
+#  define ONLY_USED_FOR_DEBUGGING(x) UNUSED(x)
 #  define MANGLE_LENGTH          0
-#  define MAYBE_UNUSED(x)           
-#  define MAYBE_UNUSED2(x,y)                     
-#  define DPRINTF0(mask, format)
-#  define DPRINTF1(mask, format,arg)
-#  define DPRINTF2(mask, format,arg1, arg2)
-#  define DPRINTF3(mask, format,arg1, arg2, arg3)
-#  define DPRINTF4(mask, format,arg1, arg2, arg3, arg4)
+#  define DPRINTF0(mask, format)  do {;} while(FALSE)
+#  define DPRINTF1(mask, format,arg) do {;} while(FALSE)
+#  define DPRINTF2(mask, format,arg1, arg2) do {;} while(FALSE)
+#  define DPRINTF3(mask, format,arg1, arg2, arg3) do {;} while(FALSE)
+#  define DPRINTF4(mask, format,arg1, arg2, arg3, arg4) do {;} while(FALSE)
+#  define DPRINTF5(mask, format,arg1, arg2, arg3, arg4, arg5) do {;} while(FALSE)
 #  define ERRMSG(b)
 #  define SHOWCURSOR
 #  define DEBUG_RANDOM_SLEEP  
