@@ -80,9 +80,13 @@ mystrndup(const char *string, int len)
 }
 
 
+/* compare strings for equality. fail if either of them is NULL */
+bool
+strings_are_equal(const char *s1, const char *s2) {
+  return s1 && s2 && strcmp(s1,s2)== 0;
+}
 
-/* mysavestring: allocate a copy of a string on the heap */
-
+/*  allocate a copy of a string on the heap */
 char *
 mysavestring(const char *string)
 {
@@ -90,6 +94,14 @@ mysavestring(const char *string)
   return mystrndup(string, strlen(string));
 }
 
+/* return "<NULL>" if arg == NULL, else arg */
+char *
+strifnull(char *string)
+{
+  return string ? string : "<NULL>";
+}       
+
+
 
 /* add3strings: allocate a sufficently long buffer on the heap and
    successively copy the three arguments into it */
@@ -137,10 +149,11 @@ append_and_free_old(char *str1, const char *str2)
 char *
 mybasename(const char *filename)
 {                               /* determine basename of "filename" */
-#if defined(HAVE_BASENAME) && defined(_GNU_SOURCE) /* we want only the GNU version */
-  char *filename_copy = mysavestring(filename);
-  char *result = basename(filename_copy);
-  return result;
+#if defined(HAVE_BASENAME) && defined(_GNU_SOURCE) /* we want only the GNU version  - but this doesn't guarantee that */
+  char *filename_copy = mysavestring(filename); 
+  char *result = mysavestring(basename(filename_copy));
+  free(filename_copy);
+  return result; /* basename on HP-UX is toxic: the result will be overwritten by subsequent invocations! */
 #else
   char *p;
 
@@ -170,7 +183,7 @@ mydirname(const char *filename)
 #endif
 }
 
-
+ 
 
 /* Better atoi() with error checking */
 int
@@ -182,7 +195,7 @@ my_atoi(const char *nptr)
   errno = 0;
   result = (int) strtol(nptr, &endptr, 10);
   if (errno || endptr == nptr || *endptr)
-    myerror(FATAL|USE_ERRNO, "Could not make sense of <%s> as an integer", mangle_string_for_debug_log(nptr, 10));
+    myerror(FATAL|USE_ERRNO, "Could not make sense of <%s> as an integer", mangle_string_for_debug_log(nptr, 20));
   return result;
 }       
 
@@ -245,6 +258,7 @@ void test_haystack(const char *haystack, const char* needle) {
 /* split_with("a bee    cee"," ") returns a pointer to an array {"a", "bee",  "cee", NULL} on the heap */
 
 char **
+
 split_with(const char *string, const char *delimiters) {
   const char *s;
   char *token, **pword;
@@ -256,11 +270,12 @@ split_with(const char *string, const char *delimiters) {
 }       
 
 /* unsplit_with(3, ["a", "bee", "cee"], "; ") returns a pointer to "a; bee; cee" on the heap */
+/* takes n elements from strings (or all of them if n < 0) */
 char *
 unsplit_with(int n, char **strings, const char *delim) {
   int i;
-  char *result = mysavestring(n> 0 ? strings[0]: "");
-  for (i = 1; i < n; i++) { 
+  char *result = mysavestring(n != 0 && strings[0] ? strings[0]: "");
+  for (i = 1; (n<0 && strings[i]) || i < n; i++) { 
        result = append_and_free_old(result, delim);
        result = append_and_free_old(result, strings[i]);
   }
@@ -268,7 +283,7 @@ unsplit_with(int n, char **strings, const char *delim) {
 }
 
 /* split_with("a\t\tbla", '\t') returns {"a" "bla", NULL}, but we want {"a", "", "bla", NULL} for filter completion.
-   We write a special version (can be freed with free_splitlist), that optionally checkst the number of components (if expected_count > 0) */
+   We write a special version (can be freed with free_splitlist), that optionally checks the number of components (if expected_count > 0) */
 char **split_on_single_char(const char *string, char c, int expected_count) {
   /* the 1st +1 for the last element ("bla"), the 2nd +1 for the marker element (NULL) */
   char **list = mymalloc((count_char_occurrences(string,c) + 1 + 1) * sizeof(char **));
@@ -286,7 +301,7 @@ char **split_on_single_char(const char *string, char c, int expected_count) {
   *pword++ = mysavestring(current_word);
   if (expected_count  && pword-list != expected_count) 
     myerror(FATAL|NOERRNO, "splitting <%s> on single %s yields %d components, expected %d",
-            mangle_string_for_debug_log(string, 30), mangle_char_for_debug_log(c, 1), pword -  list, expected_count); 
+            mangle_string_for_debug_log(string, 50), mangle_char_for_debug_log(c, 1), pword -  list, expected_count); 
   *pword = NULL;
   free(stringcopy);
   return list;
@@ -335,7 +350,7 @@ search_and_replace(char *patt, const char *repl, const char *string, int cursorp
 
   assert(patt && repl && string);
   DPRINTF2(DEBUG_READLINE, "string=%s, cursorpos=%d",
-           mangle_string_for_debug_log(string, 40), cursorpos);
+           M(string), cursorpos);
   scratchsize = max(stringlen, (stringlen * replen) / pattlen) + 1;     /* worst case : repleng > pattlen and string consists of only <patt> */
   DPRINTF1(DEBUG_READLINE, "Allocating %d bytes for scratchpad", (int) scratchsize);
   scratchpad = mymalloc(scratchsize);
@@ -432,16 +447,22 @@ mangle_string_for_debug_log(const char *string, int maxlen)
   int total_length;
   char *mangled_char, *result;
   const char *p; /* good old K&R-style *p. I have become a fossil... */
-
+  MBSTATE st;
+  
   if (!string)
     return mysavestring("(null)");
   result = mysavestring("");
-  for(p = string, total_length = 0; *p; p++) {
-    mangled_char = mangle_char_for_debug_log(*p, FALSE);
-    total_length +=  strlen(mangled_char);
+  for(mbc_initstate(&st), p = string, total_length = 0; *p;  mbc_inc(&p, &st)) {
+    if (is_multibyte(p, &st)) {
+        mangled_char = mbc_first(p, &st);
+        total_length += 1;
+      } else {
+        mangled_char = mangle_char_for_debug_log(*p, FALSE);
+        total_length +=  strlen(mangled_char); /* can be more than 1 ("CTRL-A") */
+      } 
     if (maxlen && (total_length > maxlen)) {
       result = append_and_free_old(result, "...");      
-      break;  /* break *before* we reach maxlen */
+      break;  /* break *before* we append the latest char and exceed maxlen */
     }
     result = append_and_free_old(result, mangled_char);
     free(mangled_char);
@@ -517,8 +538,9 @@ list4 (char *el0, char *el1, char *el2, char *el3)
    We need to call this function on everything we get from the
    inferior command because (out of sheer programmer laziness) rlwrap
    uses C strings internally. Zero bytes are only ever used as padding
-   (@@@is this true?), and padding is not used anymore on modern
-   terminals. (except maybe for things like the visual bell) */
+   (@@@ I have never seen this happen, by the way), and padding is not
+   used anymore on modern terminals. (except maybe for things like the
+   visual bell) */
  
 
 void remove_padding_and_terminate(char *buf, int length) {
@@ -530,7 +552,7 @@ void remove_padding_and_terminate(char *buf, int length) {
   }
   *copyptr = '\0';
   if (debug && strlen(buf) != (unsigned int) length)
-    DPRINTF2(DEBUG_TERMIO, "removed %d zero bytes (padding?) from %s", length - (int) strlen(buf), mangle_string_for_debug_log(buf, MANGLE_LENGTH));
+    DPRINTF2(DEBUG_TERMIO, "removed %d zero bytes (padding?) from %s", length - (int) strlen(buf), M(buf));
 }       
         
 #define ESCAPE  '\033'
@@ -544,7 +566,7 @@ void remove_padding_and_terminate(char *buf, int length) {
    than the original, we need not worry about writing outside buf
 
    Example: if buf="abc\bd\r\e" then, after calling unbackspace(buf),
-   buf will contain "ebd"
+   buf will contain "ebd" @@@ Should be just "e"
 
    We need this function because many commands emit "status lines"
    using backspaces and carriage returns to re-write parts of the line in-place.
@@ -553,11 +575,11 @@ void remove_padding_and_terminate(char *buf, int length) {
 */
 
 void
-unbackspace(char* buf) {
+unbackspace_old(char* buf) {
   char *readptr, *copyptr, *endptr;
   int seen_bs_or_cr;
 
-  DPRINTF1(DEBUG_TERMIO,"unbackspace: %s", mangle_string_for_debug_log(buf, MANGLE_LENGTH));
+  DPRINTF1(DEBUG_TERMIO,"unbackspace: %s", M(buf));
   seen_bs_or_cr = FALSE;
   
   for (readptr = copyptr = endptr = buf; *readptr; readptr++) {
@@ -583,21 +605,71 @@ unbackspace(char* buf) {
   }
   *endptr = '\0';
   if (seen_bs_or_cr) 
-      DPRINTF1(DEBUG_TERMIO,"unbackspace result: %s", mangle_string_for_debug_log(buf, MANGLE_LENGTH));
+      DPRINTF1(DEBUG_TERMIO,"unbackspace result: %s", M(buf));  
+}
+
+
+
+
+void
+unbackspace(char* buf) {
+  char *readptr, *copyptr;
+  int seen_bs_or_cr;
+
+  DPRINTF1(DEBUG_TERMIO,"unbackspace: %s", M(buf));
+  seen_bs_or_cr = FALSE;
+  
+  for (readptr = copyptr = buf; *readptr; readptr++) {
+    switch (*readptr) {
+    case BACKSPACE:
+      copyptr = (copyptr > buf ? copyptr - 1 : buf);  /* cannot backspace past beginning of buf          */
+      while (*copyptr == RL_PROMPT_END_IGNORE && copyptr > buf) {    /* skip control codes  ...          */
+        while (*copyptr != RL_PROMPT_START_IGNORE && copyptr > buf)  /* but never past buffer start      */
+          copyptr--;                                                 /* e.g. with pathological "x\002\b" */
+        if (copyptr > buf)
+          copyptr--;
+      }   
+      seen_bs_or_cr = TRUE;
+      break;
+    case CARRIAGE_RETURN:
+      copyptr = buf;  
+      seen_bs_or_cr = TRUE;
+      break;
+    default:
+      *copyptr++ = *readptr;      
+      break;
+    }
+  }
+  *copyptr = '\0';
+  if (seen_bs_or_cr) 
+      DPRINTF1(DEBUG_TERMIO,"unbackspace result: %s", M(buf));
   
 }
 
 
+#ifdef UNIT_TEST
+
+static
 void test_unbackspace (const char *input, const char *expected_result) {
   char *scrap = mysavestring(input);
   unbackspace(scrap);
   if (strcmp(scrap, expected_result) != 0)
-    myerror(FATAL|NOERRNO, "unbackspace %s yielded %s, expected %s",
+    myerror(FATAL|NOERRNO, "unbackspace '%s' yielded '%s', expected '%s'",
               mangle_string_for_debug_log(input,0),
               mangle_string_for_debug_log(scrap,0),
               expected_result);
 }
 
+/*  run with: make clean ; make CFLAGS='-g -DUNIT_TEST=test'; ./rlwrap <argv> */
+TESTFUNC(test, argc, argv, stage) {
+  ONLY_AT_STAGE(TEST_AT_PROGRAM_START );
+  test_unbackspace("zx\001a\002\bq","sssssssss");
+  exit(0);
+}
+
+#endif
+
+
 
 /* Readline allows to single out character sequences that take up no
    physical screen space when displayed by bracketing them with the
@@ -635,7 +707,8 @@ mark_invisible(const char *buf)
   char *result = scratchpad;
   const char **original = &buf;
   char **copy = &scratchpad;
-
+  DPRINTF1(DEBUG_AD_HOC, "mark_invisible(%s) ...", M(buf));
+  
   if (strchr(buf, RL_PROMPT_START_IGNORE))
     return mysavestring(buf); /* "invisible" parts already marked */
     
@@ -644,6 +717,7 @@ mark_invisible(const char *buf)
     assert(*copy - scratchpad < padsize);
   }
   **copy = '\0';
+  DPRINTF1(DEBUG_AD_HOC, "mark_invisible(...) = <%s>", M(result));
   return(result);       
 }       
 
@@ -693,31 +767,35 @@ copy_next(int n, const char **original, char **copy)
 
 
 
-  
-
-/* helper function: returns the number of displayed characters (the
-   "colourless length") of str (which has its unprintable sequences
-   marked with RL_PROMPT_*_IGNORE).  Puts a copy without the
-   RL_PROMPT_*_IGNORE characters in *copy_without_ignore_markers (if
-   != NULL)
+/* helper function: returns the number of displayed characters (the "colourless length") of str (which has to have its
+   unprintable sequences marked with RL_PROMPT_*_IGNORE).  Puts a copy without the RL_PROMPT_*_IGNORE characters in
+   *copy_without_ignore_markers (if != NULL)
 */
 
 int
-colourless_strlen(const char *str, char **pcopy_without_ignore_markers, int termwidth)
+colourless_strlen(const char *str, char ** pcopy_without_ignore_markers, int UNUSED(termwidth))
 {
-  int visible = TRUE;
-  int column = 0;
-  int length = strlen(str);
-  const char *p; 
-  char *q, *copy_without_ignore_markers;
-  
+  int visible  = TRUE;
+  int length   = strlen(str);
+  int i, colourless_length, colourless_bytes;
+  const char *str_ptr, *p;
+  char *copy_ptr, *copy_without_ignore_markers;
+  MBSTATE st;
 
-  assert(termwidth >= 0);
-  q = copy_without_ignore_markers = mymalloc(length + 1);
-	
-  for(p = str; *p; p++) {
-    assert (q < copy_without_ignore_markers + length); 
-    switch (*p) {
+  typedef struct {
+    char *bytes;
+    MBSTATE state;
+  } mbchar_cell;
+  
+  mbchar_cell *cellptr, *copied_cells = mymalloc((length + 1)  * sizeof(mbchar_cell));
+    
+  /* The next loop scans str, one multi-byte character at a time, constructing a colourless copy  */
+  /* cellptr always points at the next available free cell                                        */
+  for(mbc_initstate(&st), str_ptr = str, colourless_bytes = 0, cellptr = copied_cells;
+      *str_ptr;
+      mbc_inc(&str_ptr, &st)) {
+    assert (cellptr < copied_cells + length);
+    switch (*str_ptr) {
     case RL_PROMPT_START_IGNORE:
       visible = FALSE;
       continue;
@@ -725,41 +803,73 @@ colourless_strlen(const char *str, char **pcopy_without_ignore_markers, int term
       visible = TRUE;
       continue;
     case '\r':
-      if (visible) {
-        q -= column;
-        column = 0;
-        continue;
+      if (visible) {                  /* only ever interpret CR (and Backspace) when visible (i.e. outside control sequences) */
+        for ( ; cellptr > copied_cells; cellptr--)  
+          free((cellptr-1)->bytes);   /* free all cells            */
+        mbc_initstate(&st);           /* restart with virgin state */
+        colourless_bytes = 0;
       }
-      break;
+      continue;
     case '\b':
-      if ((visible && q > copy_without_ignore_markers)) {
-        q -= 1;
-        column -= 1;
-        if (termwidth && column < 0) 
-          column += termwidth;
-        continue;
+      if ((visible && cellptr > copied_cells)) {     /* except when invisible, or at beginning of copy ... */
+        cellptr -= 1;                                /* ... reset cellptr to previous (multibyte) char     */
+        colourless_bytes -= strlen(cellptr->bytes);  
+        free(cellptr->bytes);
+        if (cellptr > copied_cells)
+          st = (cellptr -1) -> state;                /* restore corresponding shift state                  */
+        else
+          mbc_initstate(&st);                        /* or initial state, if at start of line              */
       }
-      break;
-    }   
-    if (visible) {
-      *q++ = *p;
-      column +=1;
-      if (termwidth && column >= termwidth)
-        column -= termwidth;
+      continue;
     }
-  } 
-  *q = '\0';
-  DPRINTF4(DEBUG_READLINE, "colourless_strlen(\"%s\", 0x%lx, %d) = %ld",
-           mangle_string_for_debug_log(str, MANGLE_LENGTH), (long) pcopy_without_ignore_markers,
-           termwidth,  q - copy_without_ignore_markers);  
+    if (visible) {
+      MBSTATE st_copy   = st;
+      int nbytes        = mbc_charwidth(str_ptr, &st_copy);
+      char *q           = cellptr -> bytes = mymalloc(1 + nbytes);
+
+      colourless_bytes += nbytes;
+      mbc_copy(str_ptr,&q, &st); /* copy the possibly multi-byte character at str_ptr to cellptr -> bytes , incrementing q to just past the copy */
+      *q                = '\0';
+      cellptr -> state  = st; /* remember shift state after reading str_ptr, just in case a backspace comes along later */
+      cellptr           += 1;
+    }
+  } /* end of for loop */
+
+  colourless_length = cellptr - copied_cells;
+  
+  copy_without_ignore_markers = mymalloc(colourless_bytes + 1);
+  for (cellptr = copied_cells, copy_ptr = copy_without_ignore_markers, i = 0; i < colourless_length; i++, cellptr++) {
+    for(p = cellptr->bytes; *p; p++, copy_ptr++) 
+      *copy_ptr = *p;
+    free(cellptr->bytes);
+  }
+  *copy_ptr = '\0';
+  free(copied_cells);
+
+
+  DPRINTF4(DEBUG_READLINE, "colourless_strlen(\"%s\", \"%s\") = %d  chars, %d  bytes",
+           M(str), copy_without_ignore_markers, colourless_length, colourless_bytes);
+
   if (pcopy_without_ignore_markers)
     *pcopy_without_ignore_markers = copy_without_ignore_markers;
   else
-    free (copy_without_ignore_markers);
+    free(copy_without_ignore_markers);
+  
+  
 
-  return q - copy_without_ignore_markers;
+  return colourless_length;
 }
 
+/* helper function: returns the number of displayed characters (the
+   "colourless length") of str (which has its unprintable sequences
+   marked with RL_PROMPT_*_IGNORE).
+
+   Until rlwrap 0.44, this function didn't take wide characters into 
+   consideration, causing problems with long prompts containing wide characters.
+*/
+
+
+
 int
 colourless_strlen_unmarked (const char *str, int termwidth)
 {
@@ -847,56 +957,6 @@ colour_name_to_ansi_code(const char *colour_name) {
 
 
 
-/*
-  returns TRUE if 'string' matches the 'regexp' (or is a superstring
-  of it, when we don't HAVE_REGEX_H). The regexp is recompiled with
-  every call, which doesn't really hurt as this function is not called
-  often: at most twice for every prompt.  'string' and 'regexp' may be
-  NULL (in which case FALSE is returned)
-
-  Only used for the --forget-regexp and the --prompt-regexp options
-*/
-int match_regexp (const char *string, const char *regexp, int case_insensitive) {
-  int result;
-  
-  if (!regexp || !string)
-    return FALSE;
-  
-#ifndef HAVE_REGEX_H
-  {
-    static int been_warned = 0;
-    char *metachars = "*()+?";
-    char *lc_string = (case_insensitive ? lowercase(string) : mysavestring(string));
-    char *lc_regexp = (case_insensitive  ? lowercase(regexp) : mysavestring(regexp));
-
-    if (scan_metacharacters(regexp, metachars) && !been_warned++) /* warn only once if the user specifies a metacharacter */
-      myerror(WARNING|NOERRNO, "one of the regexp metacharacters \"%s\" occurs in regexp(?) \"%s\"\n"
-             "  ...but on your platform, regexp matching is not supported!", metachars, regexp);        
-    
-    result = mystrstr(lc_string, lc_regexp);
-    free(lc_string);
-    free(lc_regexp);    
-  }
-#else
-  {
-    regex_t *compiled_regexp = mymalloc(sizeof(regex_t));
-    int compile_error = regcomp(compiled_regexp, regexp, REG_EXTENDED|REG_NOSUB|(case_insensitive ? REG_ICASE : 0));
-     
-    if (compile_error) {
-      int size = regerror(compile_error, compiled_regexp, NULL, 0);
-      char *error_message =  mymalloc(size);
-      regerror(compile_error, compiled_regexp, error_message, size);
-      myerror(FATAL|NOERRNO, "in regexp \"%s\": %s", regexp, error_message);  
-    } else {
-      result = !regexec(compiled_regexp, string, 0, NULL, 0);
-      free(compiled_regexp);
-    }
-  }
-#endif
-
-  
-  return result;        
-}
 
 
 /* returns TRUE if string is numeric (i.e. positive or negative integer), otherwise FALSE */
@@ -910,10 +970,10 @@ int isnumeric(char *string){
   return TRUE;
 }
 
-#define DIGITS_NUMBER 8  /* number of digits of length of field */
-#define MAX_FIELD_LENGTH (~((-1)<<(DIGITS_NUMBER * 4 -1)) -1)
-#define MY_HEX_FORMAT(n) "%0" MY_ITOA(n) "x"
-#define MY_ITOA(n) #n
+#define DIGITS_NUMBER 8  /* number of (hex) digits of length of field. 6 digits -> max 16MB per message field, should suffice */
+#define MAX_FIELD_LENGTH ((1UL <<  (DIGITS_NUMBER * 4 - 1)) -1) /* max integer that can be written with DIGITS_NUMBER (hex)digits */
+#define MY_HEX_FORMAT(n) ("%0" MY_ITOA(n) "x")
+#define MY_ITOA(n) #n 
 
 
 /* fussy strtol with error checking */
@@ -977,10 +1037,10 @@ message can be empty, or, equivalently, NULL.
 char *
 append_field_and_free_old(char *message, const char *field)
 {
-  int length = strlen(field);
+  long unsigned int length = strlen(field);
   char *encoded_length;
   if (length > MAX_FIELD_LENGTH)
-    myerror(FATAL|NOERRNO, "message field\"%s...\" has length %d, it should be less than %d",
+    myerror(FATAL|NOERRNO, "message field\"%s...\" has length %ld, it should be less than %ld",
             mangle_string_for_debug_log(field, 10), length,  MAX_FIELD_LENGTH);
   encoded_length =  encode_field_length(length);
   message = append_and_free_old(message, encoded_length);
@@ -1054,3 +1114,282 @@ split_filter_message(char *message, int *counter)
   *plist = 0;
   return list;
 }
+
+
+
+#ifndef HAVE_REGEX_H
+char *protect_or_cleanup(const char *prompt) {
+  return mysavestring(prompt); /*essentially a NOP */
+}       
+
+#else
+
+/* regcomp with error checking (and simpler signature) */
+static regex_t *my_regcomp(const char*regex, int flags) {
+  regex_t *compiled_regexp;
+  int compile_error;
+  if (!*regex)
+    return NULL;
+  compiled_regexp = mymalloc(sizeof(regex_t));
+  compile_error = regcomp(compiled_regexp, regex, flags);     
+  if (compile_error) {
+    int size = regerror(compile_error, compiled_regexp, NULL, 0);
+    char *error_message =  mymalloc(size);
+    regerror(compile_error, compiled_regexp, error_message, size);
+    myerror(FATAL|NOERRNO, "(Internal error:) in regexp \"%s\": %s", mangle_string_for_debug_log(regex,256), error_message);  
+  }
+  return compiled_regexp;
+}       
+ 
+
+#define TOKEN '@'
+#define MAXGROUPS 2
+
+/* protect_pattern("foo", 'a', 'b) = "(a[^b]*b)|(foo)" */
+char *protect(const char *pattern, char protect_start, char protect_end) {
+  char *alternative_pattern = mymalloc(strlen(pattern) + 14);
+  sprintf(alternative_pattern,"(%c[^%c]*%c)|(%s)", protect_start, protect_end, protect_end, pattern); 
+  return alternative_pattern;
+}
+
+
+
+/* Substitute all occurences of the second group in a <compiled_pattern> "(..)|(..)" within <source> with <replacement>
+   (any TOKENs within the replacement will be replaced by the match) , skipping everything that matches the first group
+   (which can be said to "protect" against replacement) E.g: 
+        
+       replace_special("a zzz b zzz", "(a[^a]*b)|(z+)", "(@)") = "a zzz b (zzz)" 
+
+   The role of a and b above will be played by the  RL_PROMPT_{START,END}_IGNORE characters                           */
+
+char *replace_special(const char *source, regex_t *compiled_pattern, const char*replacement) {
+  const char *source_cursor;
+  char *copy_with_replacements, *copy_cursor;
+  int max_copylen;
+
+  assert(source != NULL);
+  max_copylen = 1 + max(1, strlen(replacement)) * strlen(source);
+
+  if (!compiled_pattern) /* pattern == NULL: just return a copy */
+    return mysavestring(source);
+  copy_with_replacements = mymalloc(max_copylen); /* worst case: replace every char in source by replacement (+ 1 final zero byte)  */
+  source_cursor = source;
+  copy_cursor = copy_with_replacements;
+    
+  while(TRUE) {
+    regmatch_t matches[MAXGROUPS + 1];          /* whole match + MAXGROUPS groups */
+    if ((regexec(compiled_pattern, source_cursor, MAXGROUPS + 1, matches , 0) == REG_NOMATCH)) {   /* no (more) matches ...   */
+      strcpy(copy_cursor,source_cursor);        /* .. copy remainder of source (may be empty), and we're done      */
+      break;                                    /* 0-terminates copy, even if last match consumed source copletely */
+    } else {
+      int i; const char *p;
+      int protected_start   = matches[1].rm_so;
+      int protected_end     = matches[1].rm_eo;
+      int protected_length  = protected_end - protected_start;
+      int match_start       = matches[2].rm_so;  
+      int match_end         = matches[2].rm_eo;
+      int match_length      = match_end - match_start;   /* Either the first ("protected") group in alternative_pattern matches, */
+      assert(!(protected_end > -1 && match_end > -1));   /*  ... or the second (never both - that is the assertion here)         */
+      assert(protected_length > 0 || match_length > 0);  /*  ... and that match cannot be empty                                  */
+      if (protected_end > -1)                   /* If it is the first ...                                                        */
+        for (i = 0; i< protected_end; i++)      /* copy until protected match ends                                               */
+          *copy_cursor++ =  *source_cursor++;
+      else {                                    /* if it is the second (the original pattern) ...    */
+        for (i = 0; i< match_start; i++)        /* ... copy until match starts                       */
+          *copy_cursor++ =  *source_cursor++;
+        for(p = replacement; *p; p++)           /* splice replacement                                */
+          if (*p == TOKEN) {                    /* where there is a TOKEN (may be more than one) ... */
+            assert(copy_cursor + match_length - copy_with_replacements < max_copylen);
+            for(i=0; i< match_length ; i++)     /* ... splice matched group                          */
+              *copy_cursor++ = *(source_cursor + i);
+          } else                                  /* otherwise, just copy replacement                  */
+            *copy_cursor++ = *p;
+        source_cursor += match_length;          /* finished replacing this match, try again          */
+      }
+    }
+    *copy_cursor = '\0';
+  }
+  return copy_with_replacements;
+}
+      
+
+/* All the codes we want to preserve (i.e. keep and put between RL_PROMPT_{START,END}_IGNORE )    */
+static
+char *protected_codes[]  = { "\x1B\x1B", NULL};
+
+/* As we don't protect (and hence will get rid of) ANSI colour codes when "bleaching" the prompt, */
+/* specify separately:                                                                            */ 
+static char *ansi_colour_code_regexp = "(\x1B\\[[0-9;]*m)";          /* colour codes              */
+
+
+
+/* All the codes we want to get rid of. cf.                                                                        */
+/* https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python */
+/* We only cover 7-bit C1 ANSI sequences; the 8-bit are not much used as they interfere with UTF-8, and we         */
+/* don't need to absolutely catch everything anyway                                                                */
+static
+char *unwanted_codes[] = { "(\x1B[ -/]*[0-Z\\\\-~])"           /* ANSI X3.41: ESC + I-pattern + F-pattern (except [, which is covered below) */
+                           ,"(\x1B[@-Z\\]-_])"                 /* C1_Fe */
+                           ,"(\x1B\\[[0-9:;<>=?]*[-/]*[@-~])"  /* CSI   (overlaps with "protected" codes!) (is the meaning of [@-~] locale-dependent?) */
+                           ,"(\x1B\\[?1h\x1B=)"                /* smkx keypad            */
+                           ,"(\x1B\\]0;[[:print:]]*\x07)"      /* tsl <window title> fsl */
+
+                         ,NULL};
+
+
+
+
+/* mark protected codes between RL_PROMPT_{START,END}_IGNORE and erase unwanted codes */
+char *protect_or_cleanup(char *prompt, bool free_prompt) {
+  char *result1, *result;
+  static char *protected_codes_regexp ;
+  static regex_t *compiled_and_protected_protected_codes_regexp;
+  static char *protected_token;
+  static char *unwanted_codes_regexp;
+  static regex_t *compiled_and_protected_unwanted_codes_regexp;
+  
+  /* protect stuff we want to keep: */
+  
+  /* (once)construct the regexp for the protected codes: */
+  if (!protected_codes_regexp) {
+    protected_codes_regexp = unsplit_with(-1, &protected_codes[0], "|");
+    if (!bleach_the_prompt)
+      protected_codes_regexp = add3strings(protected_codes_regexp, "|", ansi_colour_code_regexp);
+    compiled_and_protected_protected_codes_regexp = my_regcomp(protect(protected_codes_regexp, RL_PROMPT_START_IGNORE, RL_PROMPT_END_IGNORE), REG_EXTENDED);
+  }
+  /* (once) construct the replacement pattern */
+  if (!protected_token) {
+     protected_token = mymalloc(4); /* "\x01@\x02" */
+     sprintf(protected_token,"%c%c%c",RL_PROMPT_START_IGNORE, TOKEN, RL_PROMPT_END_IGNORE);
+  }
+  result1 = replace_special(prompt, compiled_and_protected_protected_codes_regexp, protected_token);
+  
+  if (!unwanted_codes_regexp) {
+    unwanted_codes_regexp = unsplit_with(-1, &unwanted_codes[0], "|");
+    DPRINTF1(DEBUG_AD_HOC, "unwanted_codes_regexp: %s", M(unwanted_codes_regexp));
+    compiled_and_protected_unwanted_codes_regexp =  my_regcomp(protect(unwanted_codes_regexp, RL_PROMPT_START_IGNORE, RL_PROMPT_END_IGNORE), REG_EXTENDED);
+  }
+  result = replace_special(result1, compiled_and_protected_unwanted_codes_regexp, "");
+   DPRINTF2(DEBUG_READLINE, "protect_or_cleanup(%s) = %s", M(prompt), M(result)); 
+  free(result1);
+  
+  if (free_prompt)
+    free(prompt);
+
+  return result;
+}
+
+
+/*
+  returns TRUE if 'string' matches the 'regexp' (or is a superstring
+  of it, when we don't HAVE_REGEX_H). The regexp is recompiled with
+  every call, which doesn't really hurt as this function is not called
+  often: at most twice for every prompt.  'string' and 'regexp' may be
+  NULL (in which case FALSE is returned)
+
+  Only used for the --forget-regexp and the --prompt-regexp options
+*/
+int match_regexp (const char *string, const char *regexp, int case_insensitive) {
+  int result = FALSE;
+  
+  if (!regexp || !string)
+    return FALSE;
+
+#ifndef HAVE_REGEX_H
+  {
+    static int been_warned = 0;
+    char *metachars = "*()+?";
+    char *lc_string = (case_insensitive ? lowercase(string) : mysavestring(string));
+    char *lc_regexp = (case_insensitive  ? lowercase(regexp) : mysavestring(regexp));
+
+    if (scan_metacharacters(regexp, metachars) && !been_warned++) /* warn only once if the user specifies a metacharacter */
+      myerror(WARNING|NOERRNO, "one of the regexp metacharacters \"%s\" occurs in regexp(?) \"%s\"\n"
+             "  ...but on your platform, regexp matching is not supported!", metachars, regexp);        
+    
+    result = mystrstr(lc_string, lc_regexp);
+    free(lc_string);
+    free(lc_regexp);    
+  }
+#else
+  {
+    regex_t *compiled_regexp = my_regcomp(regexp, REG_EXTENDED|REG_NOSUB|(case_insensitive ? REG_ICASE : 0));
+    result = !regexec(compiled_regexp, string, 0, NULL, 0);
+    regfree(compiled_regexp);
+  }
+#endif
+
+  
+  return result;        
+}
+
+
+  
+#ifdef UNIT_TEST  
+
+  
+TESTFUNC(test_subst, argc, argv, stage) {
+  ONLY_AT_STAGE(TEST_AFTER_OPTION_PARSING);
+  while(TRUE) {
+     regmatch_t matches[MAXGROUPS + 1];          /* whole match + MAXGROUPS groups */
+     char *line_read = readline ("go ahead (string pattern): ");
+     if (strings_are_equal(line_read, "stop"))
+       break;
+     char**components = split_with(line_read," ");
+     if (!(components[0] && components[1]))
+       continue;
+     regex_t *re = my_regcomp(protect(components[1],'a','b'), REG_EXTENDED);
+     bool does_match = regexec(re, components[0], MAXGROUPS + 1, matches , 0) != REG_NOMATCH;
+     printf("%s\n", does_match  ? "JA" : "NEE");
+     free_foreign(line_read);
+  }
+  cleanup_rlwrap_and_exit(EXIT_SUCCESS);
+}
+
+#endif
+
+#endif /* def HAVE_REGEX_H */
+
+
+/* scan blocks of client output for "cupcodes" that enter and exit the "alternate screen" and set the global variable screen_is_alternate accordingly */
+void check_cupcodes(const char *client_output) {
+  static char *still_unchecked; /* to avoid missing a cupcode that spans more than 1 read buffer, we keep the last few 
+                                   bytes in a static buffer (still_unchecked) , that we always prepend to the next incoming block */
+  static int rmcup_len, smcup_len, max_cuplen;
+  char *output_copy, *outputptr;
+  int cut_here;
+  
+  if (!(commands_children_not_wrapped && term_smcup && term_rmcup))
+    return; /* check is impossible or unnecessary */
+
+  if (still_unchecked == NULL) { /* init static vars once */
+    still_unchecked = mysavestring("");
+    rmcup_len = strlen(term_rmcup);
+    smcup_len = strlen(term_smcup);
+    max_cuplen = max(rmcup_len, smcup_len);
+  }     
+  outputptr = output_copy = append_and_free_old(still_unchecked, client_output);
+
+  /* keep the very end for next time, as it might contain a partial cupcode */
+  /* in the worst case we will notice the shortest of the two cupcodes twice */
+  cut_here  = max(0, strlen(output_copy) - max_cuplen);
+  still_unchecked = mysavestring (&output_copy[cut_here]);
+
+
+  while (TRUE) {
+    char *rmptr = strstr(outputptr, term_rmcup);
+    char *smptr = strstr(outputptr, term_smcup);
+    if (rmptr == smptr) {
+      assert(rmptr == NULL); /* can only fail  if term_smcup is prefix of term_rmcup, or vice versa, which never happens as far as I know */ 
+      break;
+    } else if (rmptr > smptr) { /* rmcup and smcup may occur in the same block of output */
+      DPRINTF0(DEBUG_READLINE, "Saw rmcup");
+      screen_is_alternate = FALSE;
+      outputptr = rmptr + rmcup_len; /* 1 past match */
+    } else  {
+      DPRINTF0(DEBUG_READLINE, "Saw smcup");
+      screen_is_alternate = TRUE;
+      outputptr = smptr + smcup_len;
+    }       
+  }
+  free(output_copy);
+}
