@@ -30,18 +30,22 @@ In the future it will also use terminfo or even ncurses. But the fall-back to ol
 to be especially needed on older machines                                                                                            */ 
 
 /*global vars */
-char term_eof;                  /* end_of_file char */
-char term_stop;                 /* stop (suspend) key */
-char *term_backspace;           /* backspace control seq  (or 0, if none defined in terminfo) */
-char *term_cursor_hpos;         /* control seq to position cursor st given position on current line */ 
+char term_eof;                       /* end_of_file char */
+char term_stop;                      /* stop (suspend) key */
+char *term_backspace;                /* backspace control seq  (or 0, if none defined in terminfo) */
+char *term_cursor_hpos;              /* control seq to position cursor st given position on current line */ 
 char *term_clear_screen;
 char *term_cursor_up;
 char *term_cursor_down;
-char *term_cursor_left;         /* only used for debugging (the SHOWCURSOR macro)                */
-char *term_cursor_right;        /* only used to emulate a missing term_cursor_hpos               */
-char *term_rmcup;               /* rmcup - char sequence to return from alternate screen         */
-char *term_rmkx;                /* rmkx - char sequence to return from keyboard application mode */
-/* int term_colors = -1;        number of colors ('colors' capability, only in terminfo)         */
+char *term_cursor_left;              /* only used for debugging (the SHOWCURSOR macro)                */
+char *term_cursor_right;             /* only used to emulate a missing term_cursor_hpos               */
+char *term_smcup;                    /* smcup - char sequence to switch to  alternate screen          */
+char *term_rmcup;                    /* rmcup - char sequence to return from alternate screen         */
+char *term_rmkx;                     /* rmkx - char sequence to return from keyboard application mode */
+char *term_enable_bracketed_paste;   /* If we write this to some terminals (xterm etc.) they  will    */
+                                     /* "bracket" pasted input with \e[200 and \e201                  */ 
+char *term_disable_bracketed_paste;
+int term_has_colours;
 
 int redisplay = 1;
 int newline_came_last = TRUE; /* used to determine whether rlwrap needs to ouptut a newline at the very end */
@@ -55,12 +59,12 @@ static char *term_clear_line;
 char *term_name;
 
 
-static char *my_tgetstr (char *id) {
+static char *my_tgetstr (char *id, const char *capability_name) {
 #ifdef HAVE_TERMCAP_H
   char *term_string_buf = (char *)mymalloc(2048), *tb = term_string_buf;
   char *stringcap = tgetstr(id, &tb); /*  rl_get_termcap(id) only gets capabilities used by readline */
   char *retval = stringcap ? mysavestring(stringcap) : NULL; 
-  DPRINTF2(DEBUG_TERMIO, "tgetstr(\"%s\") = %s", id, (stringcap ? mangle_string_for_debug_log(stringcap,20) : "NULL"));
+  DPRINTF3(DEBUG_TERMIO, "tgetstr(\"%s\") = %s (%s)", id, (stringcap ? M(stringcap) : "NULL"), capability_name);
   free(term_string_buf);
   return retval;
 #else
@@ -68,7 +72,7 @@ static char *my_tgetstr (char *id) {
 #endif
 }
 
-static char *my_tigetstr (char *tid) {
+static char *my_tigetstr (char *tid, const char *capability_name) {
 #ifdef HAVE_CURSES_H
   static int term_has_been_setup = FALSE;
   char *stringcap, *retval;
@@ -78,7 +82,7 @@ static char *my_tigetstr (char *tid) {
   }     
   stringcap = tigetstr(tid); 
   retval = stringcap ? mysavestring(stringcap) : NULL;
-  DPRINTF2(DEBUG_TERMIO, "tigetstr(\"%s\") = %s", tid, (stringcap ? mangle_string_for_debug_log(stringcap,20) : "NULL"));
+  DPRINTF3(DEBUG_TERMIO, "tigetstr(\"%s\") = %s (%s)", tid, (stringcap ? M(stringcap) : "NULL"), capability_name);
   return retval;
 #else
   return NULL;
@@ -87,12 +91,12 @@ static char *my_tigetstr (char *tid) {
 
 /* Get (copies of) escape codes ("string capabilities") by name. First try terminfo name, then termcap name  */
 static char *
-tigetstr_or_else_tgetstr(char *capname, char *tcap_code)
+tigetstr_or_else_tgetstr(char *capname, char *tcap_code, const char *capability_name)
 {
   char *retval;
-  retval = my_tigetstr(capname);
+  retval = my_tigetstr(capname, capability_name);
   if (!retval)
-    retval = my_tgetstr(tcap_code);
+    retval = my_tgetstr(tcap_code, capability_name);
   return retval;
 }
   
@@ -114,7 +118,7 @@ init_terminal(void)
   if (ioctl(STDIN_FILENO, TIOCGWINSZ, &winsize) < 0)
     myerror(FATAL|USE_ERRNO, "Could not get terminal size");
   
-  DPRINTF2(DEBUG_TERMIO, "winsize.ws_col: %d; winsize.ws_row: %d", winsize.ws_col, winsize.ws_row);
+  DPRINTF2(DEBUG_TERMIO, "winsize.ws_col (term width): %d; winsize.ws_row (term height): %d", winsize.ws_col, winsize.ws_row);
   if (winsize.ws_col == 0)
     myerror(FATAL|NOERRNO, "My terminal reports width=0 (is it emacs?)  I can't handle this, sorry!");
 
@@ -156,21 +160,32 @@ init_terminal(void)
 
   
   if (we_have_stringcaps)  { 
-    term_backspace      = tigetstr_or_else_tgetstr("cub1",   "le");
-    term_cr             = tigetstr_or_else_tgetstr("cr",     "cr");
-    term_clear_line     = tigetstr_or_else_tgetstr("el",     "ce");
-    term_clear_screen   = tigetstr_or_else_tgetstr("clear",  "cl");
-    term_cursor_hpos    = tigetstr_or_else_tgetstr("hpa",    "ch"); 
-    term_cursor_left    = tigetstr_or_else_tgetstr("cub1",   "le");
-    term_cursor_right   = tigetstr_or_else_tgetstr("cuf1",   "nd");
-    term_cursor_up      = tigetstr_or_else_tgetstr("cuu1",   "up");
-    term_cursor_down    = tigetstr_or_else_tgetstr("cud1",   "do");
+    term_backspace      = tigetstr_or_else_tgetstr("cub1",   "le", "backspace");
+    term_cr             = tigetstr_or_else_tgetstr("cr",     "cr", "carriage return");
+    term_clear_line     = tigetstr_or_else_tgetstr("el",     "ce", "clear line");
+    term_clear_screen   = tigetstr_or_else_tgetstr("clear",  "cl", "clear screen");
+    term_cursor_hpos    = tigetstr_or_else_tgetstr("hpa",    "ch", "set cursor horizontal position"); 
+    term_cursor_left    = tigetstr_or_else_tgetstr("cub1",   "le", "move cursor left");
+    term_cursor_right   = tigetstr_or_else_tgetstr("cuf1",   "nd", "move cursor right");
+    term_cursor_up      = tigetstr_or_else_tgetstr("cuu1",   "up", "move cursor up");
+    term_cursor_down    = tigetstr_or_else_tgetstr("cud1",   "do", "move cursor down");
+    term_has_colours    = tigetstr_or_else_tgetstr("initc", "Ic",  "initialise colour") ? TRUE : FALSE; 
+    
+    /* the following codes are never output by rlwrap, but used to recognize the use of the the alternate screen by the client, and 
+       to filter out "garbage" that is coming from commands that use them */
+    term_smcup          = tigetstr_or_else_tgetstr("smcup",  "ti", "switch to alternate screen"); 
+    term_rmcup          = tigetstr_or_else_tgetstr("rmcup",  "te", "exit alternate screen"); 
+    term_rmkx           = tigetstr_or_else_tgetstr("rmkx",   "ke", "leave keypad-transmit mode");
+
+    /* there is no way (yet) to determine whether a terminal knows about bracketed_paste_enabled paste - we cannot use tigetstr(). "Dumb" terminals do not, of course */
+
+    term_enable_bracketed_paste  = strings_are_equal(term_name, "dumb") ? NULL : "\033[?2004h";
+    term_disable_bracketed_paste = strings_are_equal(term_name, "dumb") ? NULL : "\033[?2004l";
 
     
-    /* the following codes are never output by rlwrap, but used to filter out "garbage" that is coming from commands that use them */ 
-    term_rmcup          = tigetstr_or_else_tgetstr("rmcup",  "te"); 
-    term_rmkx           = tigetstr_or_else_tgetstr("rmkx",   "ke");
-
+    if (!term_cursor_right) /* probably only on 'dumb' terminal */
+      term_cursor_right = " ";
+    
     /* 
     term_colors         = tigetnum("colors");
     DPRINTF1(DEBUG_TERMIO, "terminal colors: %d", term_colors); 
@@ -217,11 +232,10 @@ cursor_hpos(int col)
   if (term_cursor_hpos) {
     instantiated = tgoto(term_cursor_hpos, 0, col); /* tgoto with a command that takes one parameter: parameter goes to 2nd arg ("vertical position"). */
     assert(instantiated);
-    DPRINTF2(DEBUG_TERMIO, "tgoto(term_cursor_hpos, 0, %d) = %s", col, mangle_string_for_debug_log(instantiated, 20));
+    DPRINTF2(DEBUG_TERMIO, "tgoto(term_cursor_hpos, 0, %d) = %s", col, M(instantiated));
     tputs(instantiated, 1, my_putchar);
   } else {
     int i;
-    assert(term_cursor_right != NULL);
     cr();
     for (i = 0; i < col; i++)
       tputs(term_cursor_right, 1, my_putchar);
@@ -261,7 +275,7 @@ clear_line()
   if (term_clear_line)
     tputs(term_clear_line, 1, my_putchar);
   else {
-    spaces = (char *) mymalloc(width + 1);
+    spaces = (char *) mymalloc(width +1);
     for (i = 0, p = spaces; i < width; i++, p++)
       *p = ' ';
     *p = '\0';
@@ -322,8 +336,7 @@ void
 my_putstr(const char *string)
 {
   int string_length = strlen(string);
-  DPRINTF2(DEBUG_TERMIO,"wrote %d bytes to stdout: %s", string_length,
-           mangle_string_for_debug_log(string, MANGLE_LENGTH));
+  DPRINTF2(DEBUG_TERMIO,"wrote %d bytes to stdout: %s", string_length, M(string));
   if (string_length == 0)
     return;
   write_patiently(STDOUT_FILENO, string, string_length, "to stdout");
