@@ -173,7 +173,6 @@ write_patiently(int fd, const void *buffer, int count, const char *whither) {
 
 
 
-static int user_frustration_signals[] = {SIGHUP, SIGINT, SIGQUIT, SIGTERM, SIGALRM};
 
 /* keep  reading from fd and write into buffer until count bytes have been read.
    for uninterruptible_msec, restart when interrupted by a signal, after this,
@@ -190,8 +189,9 @@ read_patiently2 (int fd,
   int total_read =  0;
   int interruptible = FALSE;
 
+  DPRINTF2(DEBUG_AD_HOC, "read_patiently2 tries to read %d bytes (uninterruptible: %d msec)", count, uninterruptible_msec);
+
   assert (count >= 0);
-  unblock_signals(user_frustration_signals); /* allow users to use CTRL-C, but only after uninterruptible_msec */
   if (count > 0) {
     myalarm(uninterruptible_msec);
     while (1) {
@@ -243,7 +243,6 @@ write_patiently2(int fd,
   assert(count >= 0);
   if (count == 0)
     return; /* no-op */
-  unblock_signals(user_frustration_signals); /* allow impatient users to hit CTRL-C, but only after uninterruptible_msec */
   myalarm(uninterruptible_msec);        
   while(1) { 
     if((nwritten = write(fd, (char *)buffer + total_written, count - total_written)) <= 0) {
@@ -265,7 +264,6 @@ write_patiently2(int fd,
   }
   myalarm(0);
   DPRINTF2(DEBUG_AD_HOC, "write_patiently2 wrote %d bytes: %s", total_written, mangle_buffer_for_debug_log(buffer, total_written));
-  block_all_signals();
   return;
 }       
 
@@ -334,12 +332,17 @@ int open_unique_tempfile(const char *suffix, char **tmpfile_name) {
   
 
 static char*
-markup(const char*str)
+markup(const char* colour_name, const char*str)
 {
-  if (isatty(STDOUT_FILENO) && (ansi_colour_aware || colour_the_prompt))
-    return add3strings("\033[1;31m", str,"\033[0m");
-  else
-    return mysavestring(str);
+  char *result, *colour_code;
+  if (colour_name && isatty(STDOUT_FILENO) && (ansi_colour_aware || term_has_colours || colour_the_prompt)) {
+    colour_code = add3strings("\033[", colour_name_to_ansi_code(colour_name), "m");
+    result = add3strings(colour_code, str,"\033[0m");
+    free(colour_code);
+  } else {
+    result =  mysavestring(str);
+  }
+  return result;
 }       
 
 
@@ -350,7 +353,9 @@ markup(const char*str)
    USE_ERRNO: print perror(errno) after the message
 */
 
+   
 
+  
 void
 myerror(int error_flags, const char *message_format, ...)
 {
@@ -358,6 +363,8 @@ myerror(int error_flags, const char *message_format, ...)
   char contents[BUFFSIZE];
   int is_warning = !(error_flags & FATAL);
   char *warning_or_error = is_warning ? "warning: " : "error: ";
+  char *coloured_warning_or_error = markup(is_warning? "Magenta" : "Red", warning_or_error);
+  char *message_for_debug_log;
   static int warnings_given = 0;  
   char *message = add2strings(program_name, ": ");
 
@@ -366,7 +373,8 @@ myerror(int error_flags, const char *message_format, ...)
   vsnprintf(contents, sizeof(contents) - 1, message_format, ap);
   va_end(ap);
 
-  message = append_and_free_old(message, markup(warning_or_error)); 
+  message = append_and_free_old(message, coloured_warning_or_error);
+  free(coloured_warning_or_error);
   message = append_and_free_old(message, contents);
                              
   if ((error_flags & USE_ERRNO) && saved_errno) {
@@ -374,21 +382,26 @@ myerror(int error_flags, const char *message_format, ...)
     message = append_and_free_old(message, strerror(saved_errno));
   }                                
   message = append_and_free_old(message,"\n");                            
-    
   fflush(stdout);
-  DPRINTF2(DEBUG_ALL, "%s %s", warning_or_error, contents);
- 
+
+  message_for_debug_log = search_and_replace("\n", "; ", contents, 0, NULL, NULL); 
+  DPRINTF2(DEBUG_ALL, "%s %s", warning_or_error, message_for_debug_log);
+  free(message_for_debug_log);
 
   if (! (is_warning && nowarn))
     fputs(message, stderr); /* @@@ error reporting (still) uses buffered I/O */
+
   if (is_warning && !warnings_given++ && !nowarn) 
-    fputs("\nwarnings can be silenced by the --no-warnings (-n) option\n", stderr);
+    fputs("warnings can be silenced by the --no-warnings (-n) option\n", stderr);
   
   fflush(stderr);
   free(message);
   errno =  saved_errno;
 
-  if (error_flags & FATAL) { 
+  if (error_flags & FATAL) {
+    #ifdef DUMP_CORE_ON_ERROR
+      KA_BOOM;
+    #endif
     if (!i_am_child)
       cleanup_rlwrap_and_exit(EXIT_FAILURE);
     else /* child: die and let parent clean up */
@@ -397,21 +410,31 @@ myerror(int error_flags, const char *message_format, ...)
   
 }
 
+/* fopen with error handling. Will close and re-open if *fp != NULL; */
+void my_fopen(FILE  **pfp, const char *path, const char *mode, const char *description ) {
+  char *what = "open";
+  char *how = mode[0] == 'w' ? "writing" : mode[0] == 'a' ? "appending" : "reading";
+  if (*pfp) {
+    fclose(*pfp);
+    what = "re-open";
+  }   
+  *pfp = fopen(path, mode);
+  if (!*pfp)
+    myerror(FATAL|USE_ERRNO, "Cannot %s  %s %s for %s", what, description, path, how);
+}
 
-    
 
 void
 open_logfile(const char *filename)
 {
   time_t now;
 
-  log_fp = fopen(filename, "a");
-  if (!log_fp)
-    myerror(FATAL|USE_ERRNO, "Cannot write to logfile %s", filename);
+  my_fopen(&log_fp, filename, "a", "logfile");
   now = time(NULL);
   fprintf(log_fp, "\n\n[rlwrap] %s\n", ctime(&now));
 }
 
+
 void
 write_logfile(const char *str)
 {
@@ -430,6 +453,8 @@ filesize(const char *filename)
   return (size_t) buf.st_size;
 }
 
+
+
 void
 close_logfile()
 {
@@ -498,25 +523,11 @@ int killed_by(int status) {
   return 0;
 }       
 
-/* change_working_directory() changes rlwrap's working directory to /proc/<command_pid>/cwd
-   (on systems where this makes sense, like linux and Solaris) */
-/* TODO: clean up this mess: 
-
-   * use only a global char *slaves_working_directory (not proc_pid_cwd)
-   * factor out a function int get_new_slave_cwd(char **cwd) that returns 1 (and updates slaves_working_directory) only 
-     if the working dir has changed, and use it like this:
-
-    void change_working_directory() { 
-         if(get_new_slave_cwd(&slaves_working_directory)) 
-           chdir(slaves_working_directory);
-    }  
-*/ 
 
 
-
-/* get_new_slave_cwd(&cwd)  finds the rlwrapped command's current working directory. If this differs from cwd, free(*cwd) and
-   set *cwd to (a copy of) the new working directory. Return value: 0, or 1 if rlwrap needs to do a chdir(cwd) to again have 
-   the same working dir as the rlwrapped command.
+/* get_new_slave_cwd(&cwd) tries to find the rlwrapped command's current working directory. If this
+   differs from cwd, free(*cwd) and set *cwd to (a copy of) the new working directory. Return value: 0, or 1
+   if rlwrap needs to do a chdir(cwd) to again have the same working dir as the rlwrapped command.
 */
 static int
 get_new_slave_cwd(char **cwd)
@@ -586,7 +597,7 @@ get_new_slave_cwd(char **cwd)
  
 
 
-
+/* change_working_directory() tries to change rlwrap's working directory to the rlwrapped command's current working directory   */
 void
 change_working_directory()
 {
@@ -785,6 +796,9 @@ mymalloc(size_t size)
   ptr = malloc(size);
   if (ptr == NULL) {
     /* don't call myerror(), as this calls mymalloc() again */
+    #ifdef DUMP_CORE_ON_ERROR
+       KA_BOOM;
+    #endif
     fprintf(stderr, "Out of memory: tried in vain to allocate %d bytes\n", (int) size);
     exit(EXIT_FAILURE);
   }     
@@ -871,7 +885,7 @@ void mirror_args(pid_t command_pid) {
   long cmdline_length;
   static int been_warned = 0;
 
-  if (!stored_cmdline_filename)
+  if (!stored_cmdline_filename || !command_pid) /* uninitialized, unborn or dead command */
     return;
   if (!*stored_cmdline_filename) 
      snprintf2(stored_cmdline_filename, MAXPATHLEN , "%s/%d/cmdline", PROC_MOUNTPOINT, command_pid);
